//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ Arbol.h
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#pragma once

#include "Movimiento.h"
#include "ListaMovimientos.h"
#include "Historial.h"

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ Declaracion previa
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class Tablero;

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ CONSTANTES
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static const int MAX_PLY = 128;

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ class PV
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class PV {
public:
    // Constructor
    PV::PV();

	// Actualiza la PV con un movimiento
	void Actualizar(int ply, Movimiento mov);

	// Escribe la PV en un string
	string Obtener(void) const;

    // Setea un nuevo ply para la PV
    void NuevoPly(int ply);

    // Devuelve el movimiento de la PV para el ply elegido
    Movimiento GetMovimiento(int ply) const;

    // Limpia la PV
	void Limpiar();

private:
    Movimiento pv[MAX_PLY][MAX_PLY];
    int longitudPV[MAX_PLY];
};

// Constructor
inline PV::PV() {
    Limpiar();
}

// Actualiza la PV con un movimiento
inline void PV::Actualizar(int ply, Movimiento mov) {
	pv[ply][ply] = mov;
	for (int j = ply + 1; j < longitudPV[ply + 1]; j++)
        pv[ply][j] = pv[ply+1][j];
	longitudPV[ply] = longitudPV[ply+1];
}

// Setea un nuevo ply para la pv
inline void PV::NuevoPly(int ply) {
    longitudPV[ply] = ply;
}

// Devuelve el movimiento de la pv para el ply elegido
inline Movimiento PV::GetMovimiento(int ply) const {
    assert(ply >= 0 && ply < MAX_PLY);
    return pv[ply][ply];
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ class Arbol
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class Arbol {
public:
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ CONSTRUCTOR, DESTRUCTOR
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    Arbol(const Tablero *t);
    ~Arbol(void);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ AMISTAD
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    friend class Busqueda;

     //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ HISTORIA Y KILLERS
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Actualiza la heurstica de la historia
	void ActualizarHH(int profundidad, Movimiento mov);

    // Actualiza los killers
    void ActualizarKillers(int ply, Movimiento mov);
    
    // Limpia la heurstica de la historia
    void LimpiarHH(void);

    // Limpia los killers
    void LimpiarKillers(void);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ HISTORIA Y KILLERS
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Devuelve la heuristica de historia para la pieza en la posicion elegida
    int GetHH(Pieza pieza, Casilla casilla) const;

    // Devuelve el killer1 para el ply elegido
    Movimiento GetKiller1(int ply) const;

    // Devuelve el killer2 para el ply elegido
    Movimiento GetKiller2(int ply) const;

    // Devuelve el movimiento de la pv para el ply elegido
    Movimiento GetMovPV(int ply) const;

    // Devuelve el movimiento de la hash para el ply elegido
    Movimiento GetMovHash(int ply) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ VARIANTE PRINCIPAL
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    PV& GetVariantePrincipal(void);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ PROFUNDIDAD SELECTIVA
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    int GetProfundidadSelectiva(void) const;
    void SetProfundidadSelectiva(int prof);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ CONTEO NODOS RAIZ
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Limpia el contador de nodos en raz
    void LimpiarNodosRaiz(void);
    
    EnteroU64 GetNumeroNodosRaiz(int numeroMovimiento) const;
    void SetNumeroNodosRaiz(int numeroMovimiento, EnteroU64 nodos);

private:
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ CONSTANTES
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    static const int MAX_PLY = 128;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ DATOS
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    bool enjaque[MAX_PLY];
    Movimiento killer1[MAX_PLY];
    Movimiento killer2[MAX_PLY];
    PV variantePrincipal;
    Movimiento movimientoHash[MAX_PLY];
    int heuristicaHistoria[MAX_PIEZAS][64]; // Shared
    ListaMovimientos *listaMovimientos;
    Historial historial;
    int profundidadSelectiva;
    EnteroU64 numeroNodosRaiz[ListaMovimientos::MAX_MOVIMIENTOS];
};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ FUNCIONES INLINE
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Devuelve la heuristica de historia para la pieza en la posicion elegida
inline int Arbol::GetHH(Pieza pieza, Casilla casilla) const {
    assert(PiezaValida(pieza));
    assert(CasillaValida(casilla));
    return heuristicaHistoria[pieza][casilla];
}

// Devuelve el killer1 para el ply elegido
inline Movimiento Arbol::GetKiller1(int ply) const {
    assert(ply >= 0 && ply < MAX_PLY);
    return killer1[ply];
}

// Devuelve el killer2 para el ply elegido
inline Movimiento Arbol::GetKiller2(int ply) const {
    assert(ply >= 0 && ply < MAX_PLY);
    return killer2[ply];
}

// Devuelve el movimiento de la hash para el ply elegido
inline Movimiento Arbol::GetMovHash(int ply) const {
    assert(ply >= 0 && ply < MAX_PLY);
    return movimientoHash[ply];
}        

// Devuelve la variante principal
inline PV& Arbol::GetVariantePrincipal(void) {
    return variantePrincipal;
}

// Devuelve la profundidad selectiva
inline int Arbol::GetProfundidadSelectiva(void) const {
    return profundidadSelectiva;
}

// Establece el valor de la profundidad selectiva
inline void Arbol::SetProfundidadSelectiva(int prof) {
    profundidadSelectiva = prof;
}

// Devuelve el numero de nodos generado por un movimiento en un determinado ply
inline EnteroU64 Arbol::GetNumeroNodosRaiz(int numeroMovimiento) const {
    assert(numeroMovimiento >= 0 && numeroMovimiento <= ListaMovimientos::MAX_MOVIMIENTOS);
    return numeroNodosRaiz[numeroMovimiento];
}

// Establece el numero de nodos generado por un movimiento en un determinado ply
inline void Arbol::SetNumeroNodosRaiz(int numeroMovimiento, EnteroU64 nodos) {
    assert(numeroMovimiento >= 0 && numeroMovimiento <= ListaMovimientos::MAX_MOVIMIENTOS);
    numeroNodosRaiz[numeroMovimiento] = nodos;
}
