//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ Busqueda.cpp
//+
//+ Las funciones de bsqueda del movimiento que juega el programa
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "stdafx.h"
#include "Busqueda.h"
#include "Consola.h"
#include "ControlTiempo.h"
#include "Tablero.h"
#include "Evaluador.h"
#include "Notador.h"
#include "Estadisticas.h"
#include "TablaHash.h"

using namespace Estadisticas;

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ OPCIONES DE COMPILACION
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#define USAR_NULLMOVE               // Utiliza Null-Move Pruning
#define USAR_ASPIRATION_WINDOWS     // Utiliza Aspiration Windows
//#define USAR_IID                    // Utiliza Internal Iterative Deepening
//#define USAR_PVS                    // Utiliza PVS
#define USAR_PODADO_HASH            // Utiliza el podado de la Tabla Hash

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ class Busqueda
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Constructor
Busqueda::Busqueda(Motor &m) : motor(m), arbol(&m.GetTablero()) {
    // El tablero sobre el que se calcula el movimiento
    tablero = &motor.GetTablero();
    // El evaluador de posiciones
    eval = &motor.GetEval();
    // La tabla hash
    tablaHash = &motor.GetTablaHash();
    // Resumen de bsqueda activado
#if defined(DEBUG_MOSTRAR)
    HabilitarResumen();
#else
    DeshabilitarResumen();
#endif
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ Busqueda Nivel 0
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Simula una bsqueda sencilla
Movimiento Busqueda::CalcularMejorMovimiento(void) {
    int puntaje;
    int maxProfundidad;
    int ply;

    // El tiempo de movimiento
    tiempoInicio.AsignarTiempoActual();
    tiempoBusqueda = controlTiempo.Asignar(tablero->GetEstado().GetLadoMueve());

    // Estado inicial de las variables
    detenerBusqueda = false;
    alphaRaiz = -Infinito;
    betaRaiz = +Infinito;
    ply = 1;
    LimpiarEstadisticas();
    nodosControlTiempo = NODOS_CONTROL_TIEMPO;
    arbol.GetVariantePrincipal().Limpiar();
    arbol.LimpiarKillers();
    arbol.LimpiarHH();
    arbol.LimpiarNodosRaiz();
    arbol.SetProfundidadSelectiva(0);
    tablaHash->ActualizarEdad();
    mejorMovimientoRaiz.SetNulo();

    // El valor de la mxima profundidad de bsqueda
    if(controlTiempo.GetProfundidad())
        maxProfundidad = controlTiempo.GetProfundidad() + 1;
    else
        maxProfundidad = MAX_PROFUNDIDAD; 
    
    // La lista de movimientos en raz    
    arbol.listaMovimientos[ply].SetTipoLista(COMPLETA_RAIZ);

    // Iterative Deepening: Itera el valor de profundidad para la bsqueda
    for(iteracionActual = 1; iteracionActual < maxProfundidad; iteracionActual++) {
        // Info UCI Nueva Profundidad
        GetConsola().Salida().NuevaProfundidadBusqueda(iteracionActual);

#if defined(USAR_ASPIRATION_WINDOWS)
        // Aspiration Windows
        while(true) {
            // Se realiza la bsqueda
            puntaje = BuscarRaiz(iteracionActual*PLY, alphaRaiz, betaRaiz, 1);            

            // Se acab el tiempo de bsqueda o se la interrumpi
            if(detenerBusqueda)
                break;

            // Chequeamos si el puntaje se sali de la ventana
            if(puntaje <= alphaRaiz || puntaje >= betaRaiz) {
                if(puntaje >= betaRaiz)
                    cout << "info depth " << iteracionActual << " lowerbound" << endl;
                else if(puntaje <= alphaRaiz)
                    cout << "info depth " << iteracionActual << " upperbound" << endl;

                // Se maximiza la ventana
                alphaRaiz = -Infinito;
                betaRaiz = +Infinito;
                rebusquedasAspiration++;
                continue;
            }
            else {
                // Se utiliza una ventana reducida
                alphaRaiz = max(puntaje - MARGEN_ASPIRACION, -Infinito);
                betaRaiz = min(puntaje + MARGEN_ASPIRACION, Infinito);
                break;
            }
        }
#else
        puntaje = BuscarRaiz(iteracionActual*PLY, alphaRaiz, betaRaiz, 1);
#endif

        // Se acab el tiempo de bsqueda o se la interrumpi
        if(detenerBusqueda)
            break;

        // Vemos si el puntaje es de mate y buscamos a suficiente profundidad para detener la bsqueda
        if(EsPuntajeMate(puntaje) && EsPuntajeMate(puntaje)*2 < iteracionActual)
            break;
    }
    
    // Estadisticas de bsqueda
    if(EsResumenHabilitado())
        MostrarEstadisticas();
   
    // Devolvemos el movimiento elegido
    return mejorMovimientoRaiz;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ Busqueda en raz
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int Busqueda::BuscarRaiz(int prof, int alpha, int beta, int ply) {
    Movimiento m;
    int puntaje;
    int extensiones;

    // Estado inicial de variables
    int mejorPuntaje = -Infinito;
    int movimientosBuscados = 0;
    int alphaInicial = alpha;
    bool amenaza = false;
    bool pvEncontrada = false;
    EnteroU64 conteoNodosAnterior = Estadisticas::nodosBusqueda;
    arbol.movimientoHash[ply].SetNulo();
    arbol.GetVariantePrincipal().NuevoPly(ply);

    // Tabla hash: En Raiz solo para ordenado de movimientos
    switch(tablaHash->Leer(tablero->GetHashkey(),prof)) {
        case HashAlpha:
        case HashBeta:
        case HashExacto:
        case HashInfoSecundaria:
            arbol.movimientoHash[ply] = tablaHash->GetMovimiento();
            amenaza = tablaHash->GetAmenazaMate();
    }
    
    // Extensiones jaque raz
    arbol.enjaque[ply] = tablero->EnJaque();
    if(arbol.enjaque[ply]) {
        extensionesJaque++;
        prof += EXT_JAQUE;
    }
    
    // Bucle para pasar los movimientos
    arbol.listaMovimientos[ply].Reiniciar();
    while((m = arbol.listaMovimientos[ply].GetSiguiente(ply)).EsNulo() == false) {
        // Hacer el movimiento
        tablero->Hacer(m); // La legalidad est asegurada ya que se eliminaron los ilegales
        extensiones = 0;

        // Info UCI Movimiento Actual
        if(prof >= 6*PLY) {
            GetConsola().Salida().NuevoMovimientoRaiz(m, movimientosBuscados+1, nodosBusqueda);
        }

        // Extensiones Jaque
        arbol.enjaque[ply+1] = tablero->EnJaque();
        if(arbol.enjaque[ply+1]) {
            extensiones += EXT_JAQUE;
            extensionesJaque++;
        }

        // Extensin de movimiento de pen a sptima
        if((m.GetPieza() == PeonB && Fila(m.Destino()) == Fila7) || (m.GetPieza() == PeonN && Fila(m.Destino()) == Fila2)) {
            extensiones += EXT_PEON_SEPTIMA;
            extensionesPeonSeptima++;
        }

        // Reduccin de la profundidad
        extensiones -= PLY;

        // Alpha-Beta + PVS
#if defined(USAR_PVS)
        if(!pvEncontrada || iteracionActual == 1) {
            // El primer movimiento se lo busca con una ventana completa
            pvsCompletas++;
            if(prof+extensiones >= PLY)
                puntaje = -Buscar(prof+extensiones, -beta, -alpha, ply+1, true);
            else
                puntaje = -Qsearch(-beta, -alpha, ply+1);
        }
        else {
            // Los demas movimientos se buscan con una ventana mnima
            pvsMinimas++;
            if(prof+extensiones >= PLY)
                puntaje = -Buscar(prof+extensiones, -alpha-1, -alpha, ply+1, true);
            else
                puntaje = -Qsearch(-alpha-1, -alpha, ply+1);

            // Si el puntaje est fuera de la ventana, rebuscamos con una ventana completa
            if((puntaje > alpha) && (puntaje < beta) && !detenerBusqueda) {
                pvsRebusquedas++;
                if(prof+extensiones >= PLY)
                    puntaje = -Buscar(prof+extensiones, -beta, -alpha, ply+1, true);
                else
                    puntaje = -Qsearch(-beta, -alpha, ply+1);
            }
        }
#else
        if(prof+extensiones >= PLY)
            puntaje = -Buscar(prof+extensiones, -beta, -alpha, ply+1, true);
        else
            puntaje = -Qsearch(-beta, -alpha, ply+1);
#endif

        // Deshacer el movimiento
        tablero->Deshacer();
        movimientosBuscados++;

        // Conteo nodos raz para el ordenado especial
        arbol.SetNumeroNodosRaiz(movimientosBuscados-1, Estadisticas::nodosBusqueda - conteoNodosAnterior);
        conteoNodosAnterior = Estadisticas::nodosBusqueda;

        // Se para la bsqueda
        if(detenerBusqueda)
            return 0;

        // Se encuentra un movimiento mejor
        // Esto para asegurar devolver algn movimiento por lo menos
        if(puntaje > mejorPuntaje) {
            mejorMovimientoRaiz = m;
            mejorPuntaje = puntaje;
        }

        // Alpha Beta: se mejora el movimiento actual
        if(puntaje > alpha) {
            arbol.GetVariantePrincipal().Actualizar(ply,m);
            if(m.Captura() == Vacio) {
                arbol.ActualizarKillers(ply,m);
                arbol.ActualizarHH(prof/PLY,m);
            }            
        
            alpha = puntaje;
            pvEncontrada = true;

            // Info UCI Mejor Variante Actual
            if(EsPuntajeMate(puntaje)) {
                GetConsola().Salida().NuevaVariantePrincipal(iteracionActual, arbol.GetProfundidadSelectiva(),
                    tiempoInicio.Transcurrido(), nodosBusqueda, tablaHash->GetUtilizacion(),
                    EsPuntajeMate(puntaje), true, arbol.GetVariantePrincipal().Obtener());
            }
            else {
                GetConsola().Salida().NuevaVariantePrincipal(iteracionActual, arbol.GetProfundidadSelectiva(),
                    tiempoInicio.Transcurrido(), nodosBusqueda, tablaHash->GetUtilizacion(),
                    puntaje, false, arbol.GetVariantePrincipal().Obtener());
            }
        }
    }

    // La posicin es de mate o tablas
    if(!movimientosBuscados) {
        if(arbol.enjaque[ply])
            return -Infinito+ply;
        else
            return TABLAS;
    }
    else if(tablero->GetEstado().EsReglaCincuentaAplicable()) {
        return TABLAS;
    }
    
    // Tabla Hash
    if(alphaInicial == alpha)
        tablaHash->Guardar(tablero->GetHashkey(), ply, alpha, HashAlpha, mejorMovimientoRaiz, prof, amenaza);
    else
        tablaHash->Guardar(tablero->GetHashkey(), ply, alpha, HashExacto, mejorMovimientoRaiz, prof, amenaza);
    
    return mejorPuntaje;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ Busqueda en profundidad
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int Busqueda::Buscar(int prof, int alpha, int beta, int ply, bool hacerNullMove) {
    Movimiento m;
    Movimiento mejorMov;
    int puntaje;
    
    // Incrementamos los nodos y comprobamos si debemos parar
    nodosBusqueda++;
    if(ChequeoFinBusqueda())
        return 0;

    // Maxima profundidad posible
    if(ply >= Arbol::MAX_PLY)
        return alpha;

    // Estado inicial de variables
    int extensiones = 0;
    int movimientosBuscados = 0;
    int alphaInicial = alpha;
    bool amenaza = false;
    bool pvEncontrada = false;
    mejorMov.SetNulo();
    arbol.movimientoHash[ply].SetNulo();
    arbol.GetVariantePrincipal().NuevoPly(ply);

    // Tablas por 50 movimientos
    if(tablero->GetEstado().EsReglaCincuentaAplicable()) {
        return TABLAS;
    }

    // Conteo de repeticiones
    int repeticionesBusqueda = tablero->ConteoRepeticionesEnBusqueda(ply);
    int repeticionesHistoria = tablero->ConteoRepeticionesEnHistoria(ply);
    if(ply > 2) {
        if(repeticionesBusqueda >= 2)
            return TABLAS;
        if(repeticionesBusqueda+repeticionesHistoria >= 2)
            return TABLAS;
    }
    else {
        if(repeticionesBusqueda+repeticionesHistoria >= 3)
            return TABLAS;
    }
    
    // Tabla hash
    int puntajeHash;
    switch(tablaHash->Leer(tablero->GetHashkey(),prof)) {
        case HashAlpha:
            arbol.movimientoHash[ply] = tablaHash->GetMovimiento();
            puntajeHash = tablaHash->GetPuntaje(ply);
            amenaza = tablaHash->GetAmenazaMate();
            // Vemos si debemos no hacer Null-Move en esta posicin por amenaza de mate
            if(prof-3*PLY <= tablaHash->GetProfundidad() && puntajeHash < beta)
                hacerNullMove = false;
#if defined(USAR_PODADO_HASH)
            if(puntajeHash <= alpha) {
                hashCutsAlpha++;
                return alpha;
            }
#endif
            break;
        case HashBeta:
            arbol.movimientoHash[ply] = tablaHash->GetMovimiento();
            puntajeHash = tablaHash->GetPuntaje(ply);
            amenaza = tablaHash->GetAmenazaMate();
#if defined(USAR_PODADO_HASH)
            if(puntajeHash >= beta) {
                hashCutsBeta++;
                return beta;
            }
#endif
            break;
        case HashExacto:
            arbol.movimientoHash[ply] = tablaHash->GetMovimiento();
            amenaza = tablaHash->GetAmenazaMate();
#if defined(USAR_PODADO_HASH)
            hashCutsExacto++;
            return tablaHash->GetPuntaje(ply);
#endif
            break;
        case HashInfoSecundaria:
            arbol.movimientoHash[ply] = tablaHash->GetMovimiento();
            amenaza = tablaHash->GetAmenazaMate();
            break;
    }

    // Null-Move Pruning
#if defined(USAR_NULLMOVE)
    arbol.enjaque[ply+1] = false;
    int numPiezas = tablero->GetTotalValorPiezas(tablero->GetEstado().GetLadoMueve());
    if(hacerNullMove && !amenaza && !arbol.enjaque[ply] && numPiezas && (numPiezas > 9 || prof < 7*PLY)) {
        // Profundidad del nullmove
        int profNullMove = (prof > 6*PLY && numPiezas > 9) ? NULL_MAX : NULL_MIN;
        tablero->HacerMovimientoNulo();
        nullMoveIntentos++;
        
        // Bsqueda recursiva
        if(prof-profNullMove >= PLY)
            puntaje = -Buscar(prof-profNullMove, -beta, 1-beta, ply+1, false);
        else
            puntaje = -Qsearch(-beta, 1-beta, ply+1);

        tablero->DeshacerMovimientoNulo();

        // Si se termin la bsqueda
        if(detenerBusqueda)
            return 0;

        // Cut-off
        if(puntaje >= beta) {
            if(profNullMove == NULL_MAX)
                nullMoveCutsMax++;
            else
                nullMoveCutsMin++;

            tablaHash->Guardar(tablero->GetHashkey(), ply, puntaje, HashBeta, mejorMov, prof, amenaza);
            return beta;
        }
        // Null-Move Mate Threat Extension
        else if(puntaje < MateNegativo) {
            amenaza = true;
        }
    }
#endif

    // Internal Iterative Deepening
    // Solo cuando se est siguiendo la PV
#if defined(USAR_IID)
    if(arbol.movimientoHash[ply].EsNulo() && hacerNullMove && prof >= 3*PLY) {
        // Comprobamos que estemos siguiendo la PV
        if( (ply & 1 && (alpha == alphaRaiz) && (beta == betaRaiz)) || (!(ply & 1) && (alpha == -betaRaiz) && (beta == -alphaRaiz)) ) {
            intentosIID++;
            // Bsqueda reducida
            if(prof-2*PLY >= PLY)
                puntaje = Buscar(prof-2*PLY, alpha, beta, ply, true);
            else
                puntaje = Qsearch(alpha, beta, ply);
            if(detenerBusqueda)
                return 0;
            if(puntaje <= alpha) {
                intentosIID++;
                // La bsqueda no sirve, ampliamos la ventana
                if(prof-2*PLY >= PLY)
                    puntaje = Buscar(prof-2*PLY, -Infinito, beta, ply, true);
                else
                    puntaje = Qsearch(-Infinito, beta, ply);
                if(detenerBusqueda)
                    return 0;
            }

            // Obtenemos el movimiento de la PV como movimiento de la hash
            arbol.movimientoHash[ply] = arbol.GetVariantePrincipal().GetMovimiento(ply);
        }
    }
#endif

    // Bucle de eleccin de movimiento
    if(arbol.enjaque[ply])
        arbol.listaMovimientos[ply].SetTipoLista(EVASIONES);
    else
        arbol.listaMovimientos[ply].SetTipoLista(COMPLETA);

    while((m = arbol.listaMovimientos[ply].GetSiguiente(ply)).EsNulo() == false) {
        // Hacer el movimiento
        tablero->Hacer(m);
        if(!arbol.enjaque[ply] && !tablero->PosicionLegal()) {
            tablero->Deshacer();
            continue;
        }
        else {
            extensiones = 0;
        }

        // Extensiones y reducciones
        extensiones = AplicarExtensiones(ply+1, m, amenaza);
        extensiones -= PLY;

        // Alpha-Beta + PVS
#if defined(USAR_PVS)
        if(!pvEncontrada) {
            // El primer movimiento se lo busca con una ventana completa
            pvsCompletas++;
            if(prof+extensiones >= PLY)
                puntaje = -Buscar(prof+extensiones, -beta, -alpha, ply+1, hacerNullMove);
            else
                puntaje = -Qsearch(-beta, -alpha, ply+1);
        }
        else {
            // Los demas movimientos se buscan con una ventana mnima
            pvsMinimas++;
            if(prof+extensiones >= PLY)
                puntaje = -Buscar(prof+extensiones, -alpha-1, -alpha, ply+1, hacerNullMove);
            else
                puntaje = -Qsearch(-alpha-1, -alpha, ply+1);

            // Si el puntaje est fuera de la ventana, rebuscamos con una ventana completa
            if((puntaje > alpha) && (puntaje < beta) && !detenerBusqueda) {
                pvsRebusquedas++;
                if(prof+extensiones >= PLY)
                    puntaje = -Buscar(prof+extensiones, -beta, -alpha, ply+1, hacerNullMove);
                else
                    puntaje = -Qsearch(-beta, -alpha, ply+1);
            }
        }
#else
        if(prof+extensiones >= PLY)
            puntaje = -Buscar(prof+extensiones, -beta, -alpha, ply+1, hacerNullMove);
        else
            puntaje = -Qsearch(-beta, -alpha, ply+1);
#endif

        // Deshacer el movimiento
        tablero->Deshacer();
        movimientosBuscados++;

        // Si se termin la bsqueda
        if(detenerBusqueda)
            return 0;

        // Cut-off
        if(puntaje >= beta) {
            // Estadisticas
            if(movimientosBuscados == 1)
                failHighPrimero++;
            failHigh++;

            // Tabla Hash
            tablaHash->Guardar(tablero->GetHashkey(), ply, beta, HashBeta, m, prof, amenaza);

            return beta;
        }

        // Mejoramos la posicin
        if(puntaje > alpha) {
            alpha = puntaje;
            mejorMov = m;
            pvEncontrada = true;
            arbol.GetVariantePrincipal().Actualizar(ply,m);
            if(m.Captura() == Vacio) {
                arbol.ActualizarKillers(ply,m);
                arbol.ActualizarHH(prof/PLY,m);
            }
        }
    }

    // Puntaje de mate o tablas
    if(!movimientosBuscados) {
        if(arbol.enjaque[ply])
            return -Infinito+ply;
        else
            return TABLAS;
    }

    // Tabla hash
    if(alphaInicial == alpha)
        tablaHash->Guardar(tablero->GetHashkey(), ply, alpha, HashAlpha, mejorMov, prof, amenaza);
    else
        tablaHash->Guardar(tablero->GetHashkey(), ply, alpha, HashExacto, mejorMov, prof, amenaza);
    
    return alpha;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ Quiescent Search
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int Busqueda::Qsearch(int alpha, int beta, int ply) {
    Movimiento m;
    int puntaje;
    int alphaInicial;

    // Incrementamos los nodos y comprobamos si debemos parar
    nodosBusqueda++;
    nodosQsearch++;

    // Estado inicial de variables
    arbol.GetVariantePrincipal().NuevoPly(ply);
    alphaInicial = alpha;

    // La profundidad selectiva
    if(ply > arbol.GetProfundidadSelectiva())
        arbol.SetProfundidadSelectiva(ply);

    // La evaluacin de la posicin
    puntaje = eval->Evaluar(*tablero);
    if(puntaje >= beta)
        return beta;
    if(puntaje > alpha)
        alpha = puntaje;

    // La mxima profundidad
    if(ply >= Arbol::MAX_PLY)
        return alpha;

    // Bucle de eleccin de movimiento
    arbol.listaMovimientos[ply].SetTipoLista(CAPTURAS);
    while((m = arbol.listaMovimientos[ply].GetSiguiente(ply)).EsNulo() == false) {

        // Obtenemos el puntaje de la captura, o sea, el valor del SEE
        // Si es < 0, la captura pierde material asi que se puede desechar
        if(arbol.listaMovimientos[ply].GetPuntajeUltimoMov() < 0) {
            podadosSEE++;
            continue;
        }

        // Hacer el movimiento
        tablero->Hacer(m);
        if(!tablero->PosicionLegal()) {
            tablero->Deshacer();
            continue;
        }

        // Bsqueda recursiva
        puntaje = -Qsearch(-beta,-alpha,ply+1);

        // Deshacer el movimiento
        tablero->Deshacer();

        // Alpha-Beta
        if(puntaje >= beta)
            return beta;
        if(puntaje > alpha) {
            alpha = puntaje;
            arbol.GetVariantePrincipal().Actualizar(ply,m);
        }
    }

    return alpha;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ EXTENSIONES
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Aplica extensiones a la lnea actual
int Busqueda::AplicarExtensiones(int ply, Movimiento m, bool amenaza) {
    int extensiones = 0;    
  
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ Extensiones primarias: jaque, amenaza, respuesta nica al jaque
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    
    // Extensin de jaque
    arbol.enjaque[ply] = tablero->EnJaque();
    if(arbol.enjaque[ply]) {
        extensiones += EXT_JAQUE;
        extensionesJaque++;
    }

    // Extensin amenaza mate (Null Move Mate Threat Extension)
    if(amenaza) {
        extensiones += EXT_AMENAZA_MATE;
        extensionesAmenazaMate++;
    }

    // Extensin de respuesta nica al jaque
    if(arbol.enjaque[ply-1] && arbol.listaMovimientos[ply].Cantidad() == 1) {
        extensiones += EXT_RESP_UNICA_JAQUE;
        extensionesRespuestaUnicaJaque++;
    }

    // Las extensiones principales han sido intentadas, si alguna se realiza
    // no intentamos aplicar mas extensiones y cortamos ac
    if(extensiones) {
        return LimitarExtensiones(extensiones, ply);
    }

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ Extensiones secundarias: movimiento de pen a sptima fila, recaptura,
    //+ captura a final de peones
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Extensin de movimiento de pen a sptima fila
    if((m.GetPieza() == PeonB && Fila(m.Destino()) == Fila7) || (m.GetPieza() == PeonN && Fila(m.Destino()) == Fila2)) {
        extensiones += EXT_PEON_SEPTIMA;
        extensionesPeonSeptima++;
    }

    return LimitarExtensiones(extensiones, ply);
}

// Limita las extensiones
inline int Busqueda::LimitarExtensiones(int extensiones, int ply) const {
    // Como mximo aumentamos un PLY
    return min(extensiones, PLY);
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ CONTROL DE TIEMPO - CHEQUEO FIN DE BUSQUEDA
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ControlTiempo& Busqueda::GetControlTiempo(void) {
    return controlTiempo;
}

// Chequea si se acab el tiempo de bsqueda o si el usuario la interrumpi
bool Busqueda::ChequeoFinBusqueda(void) {
    // Comprobamos el fin de la bsqueda cada x=nodosControlTiempo nodos
    if(nodosBusqueda > nodosControlTiempo) {
        nodosControlTiempo += NODOS_CONTROL_TIEMPO;

        // Se chequea si se termin el tiempo de bsqueda
        if(!controlTiempo.EsInfinito() && !controlTiempo.GetProfundidad() && tiempoInicio.Transcurrido() >= tiempoBusqueda) {
            detenerBusqueda = true;
        }

        // Se chequea si el usuario par la bsqueda
        if(motor.GetConsola().ChequearInput())
            motor.GetConsola().ProcesarInput();
    }

    return (detenerBusqueda);
}

// Para la bsqueda
void Busqueda::PararBusqueda(void) {
    detenerBusqueda = true;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ PERFT
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Calcula el perft hasta la profundidad deseada
void Busqueda::CalcularPerft(int profundidad) {
#if defined(USAR_PERFT)
    Tiempo reloj;
    double transcurrido;

    nodosPerft = 0;
    reloj.AsignarTiempoActual();
    Perft(profundidad, 1);
    transcurrido = Segundos(reloj.Transcurrido());

    GetConsola().Salida() << "Perft(" << profundidad << ") = " << nodosPerft << endl;
    GetConsola().Salida() << "Transcurrido: " << transcurrido << " seg." << endl;
    if(transcurrido) GetConsola().Salida() << "NPS: " << static_cast<EnteroU64>(nodosPerft / transcurrido) << endl;
#endif
}

// Calcula el perft-divide hasta la profundidad deseada
void Busqueda::CalcularPerftDivide(int prof) {
#if defined(USAR_PERFT)
    Movimiento m;
    int movimientos;
    EnteroU64 nodosAnterior;
    Tiempo reloj;

    reloj.AsignarTiempoActual();
    nodosPerft = nodosAnterior = movimientos = 0;
    int ply = 1;

    arbol.listaMovimientos[ply].SetTipoLista(PERFT);
    while((m = arbol.listaMovimientos[ply].GetSiguiente()).EsNulo() == false) {
        tablero->Hacer(m);
        if(tablero->PosicionLegal()) {
            GetConsola().Salida() << "[" << m << "] = ";
            if(prof-1 > 0)
                Perft(prof-1, ply+1);
            else
                nodosPerft++;
            GetConsola().Salida() << nodosPerft-nodosAnterior << endl;
            nodosAnterior = nodosPerft;
            movimientos++;
        }
        tablero->Deshacer();
    }

    GetConsola().Salida() << "Perft Divide(" << prof << ") = " << nodosPerft << endl;
    GetConsola().Salida() << "Movimientos: " << movimientos << endl;
    GetConsola().Salida() << "Transcurrido: " << Segundos(reloj.Transcurrido()) << " seg." << endl;
#endif
}

// Parte recursiva del calculo de perft
void Busqueda::Perft(int prof, int ply) {
#if defined(USAR_PERFT)
    Movimiento m;

    arbol.listaMovimientos[ply].SetTipoLista(PERFT);
    while((m = arbol.listaMovimientos[ply].GetSiguiente()).EsNulo() == false) {
        tablero->Hacer(m);
        if(tablero->PosicionLegal()) {
            if(prof-1 > 0)
                Perft(prof-1, ply+1);
            else
                nodosPerft++;
        }
        tablero->Deshacer();
    }
#endif
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ RESUMEN BSQUEDA
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Habilita el resumen de estadisticas de bsqueda al terminar
void Busqueda::HabilitarResumen(void) {
    mostrarResumen = true;
}

// Deshabilita el resumen de estadsticas de bsqueda
void Busqueda::DeshabilitarResumen(void) {
    mostrarResumen = false;
}

// Devuelve el estado actual del resumen de estadsticas
bool Busqueda::EsResumenHabilitado(void) const {
    return mostrarResumen;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ UTILIDADES
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Devuelve el numero de movimientos hasta el mate si el puntaje es de mate
inline int Busqueda::EsPuntajeMate(int puntaje) const {
    if(puntaje > MatePositivo || puntaje < MateNegativo)
        return (Infinito-abs(puntaje)+1)/2;
    return 0;
}

// Devuelve el acceso a la consola a travs del motor
inline Consola& Busqueda::GetConsola(void) const {
    return motor.GetConsola();
}

