//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ Busqueda.h
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#pragma once

#include "Movimiento.h"
#include "Tiempo.h"
#include "Motor.h"
#include "ListaMovimientos.h"
#include "Arbol.h"
#include "ControlTiempo.h"

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ DECLARACION PREVIA DE CLASES
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class TablaHash;

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ class Busqueda
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class Busqueda {
public:
    // Constructor
    Busqueda(Motor &m);

    // Devuelve el mejor movimiento para la posicin segn el motor
    Movimiento CalcularMejorMovimiento(void);

    // Para la bsqueda
    void PararBusqueda(void);

    // Devuelve el control de tiempo
    ControlTiempo &GetControlTiempo(void);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ PERFT
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Calcula el perft hasta la profundidad deseada
    void CalcularPerft(int profundidad);

    // Calcula el perft-divide hasta la profundidad deseada
    void CalcularPerftDivide(int profundidad);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ OPCIONES
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Habilita el resumen de estadisticas de bsqueda al terminar
    void HabilitarResumen(void);

    // Deshabilita el resumen de estadsticas de bsqueda
    void DeshabilitarResumen(void);

    // Devuelve el estado actual del resumen de estadsticas
    bool EsResumenHabilitado(void) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ UTILIDADES
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    static const int Infinito = 32768;
    static const int MatePositivo = (Infinito-500);
    static const int MateNegativo = (-Infinito+500);

private:
    
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ BUSQUEDA
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // La funcin recursiva que realiza casi todo el trabajo
    int Buscar(int prof, int alpha, int beta, int ply, bool hacerNullMove);

    // Realiza la bsqueda en raz
    int BuscarRaiz(int prof, int alpha, int beta, int ply);

    // Realiza la quiescent search
    int Qsearch(int alpha, int beta, int ply);

    // Chequea si se debe parar la busqueda por control de tiempo o decision del usuario
    bool ChequeoFinBusqueda(void);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ EXTENSIONES
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Aplica extensiones a la lnea actual
    int AplicarExtensiones(int ply, Movimiento m, bool amenaza);

    // Limita las extensiones segn el ply
    int LimitarExtensiones(int extensiones, int ply) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ PERFT
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Parte recursiva del calculo de perft
    void Perft(int prof, int ply);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ UTILIDADES
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Devuelve el numero de movimientos hasta el mate si el puntaje es de mate
    int EsPuntajeMate(int puntaje) const;

    // Devuelve el acceso a la consola a travs del motor
    Consola& GetConsola(void) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ CONSTANTES
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    static const int PLY = 8;

    static const int MAX_PROFUNDIDAD = 100;
    static const int NODOS_CONTROL_TIEMPO = 25000;

    static const int TABLAS = 0;

    static const int EXT_JAQUE = PLY;
    static const int EXT_AMENAZA_MATE = (3*PLY)/4;
    static const int EXT_RESP_UNICA_JAQUE = (3*PLY)/4;
    static const int EXT_PEON_SEPTIMA = (3*PLY)/4;
    static const int EXT_RECAPTURA = (3*PLY)/4;

    static const int NULL_MAX = 3*PLY;
    static const int NULL_MIN = 2*PLY;

    static const int MARGEN_FUTILITY = 151; // (ValorAlfil+1)/2
    static const int MARGEN_RAZOR = 451; // (ValorDama+1)/2

    static const int MARGEN_ASPIRACION = 50;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ ACCESO A LOS OTROS MODULOS
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    Motor &motor;
    Tablero *tablero;
    Evaluador *eval;
    TablaHash *tablaHash;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ PARA LA BUSQUEDA
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // El arbol de variantes generado en la bsqueda
    Arbol arbol;

    // El control de tiempo
    ControlTiempo controlTiempo;
    
    // Para saber cuando se debe terminar la busqueda
    bool detenerBusqueda;

    // Para calcular el tiempo de pensamiento
    Tiempo tiempoInicio;
    EnteroU64 nodosControlTiempo;
    TipoTiempo tiempoBusqueda;

    // Para guardar el movimiento que se va a devolver
    Movimiento mejorMovimientoRaiz;

    // La iteracion actual
    int iteracionActual;

    // Los valores de alpha y beta en raz
    int alphaRaiz;
    int betaRaiz;

    // Opciones
    bool mostrarResumen;
};
