//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ ConsolaUCI.cpp
//+
//+ Una consola con soporte del protocolo UCI (Universal Chess Interface)
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "stdafx.h"
#include "ConsolaUCI.h"

#include "Busqueda.h"
#include "Tablero.h"
#include "ControlTiempo.h"
#include "TablaHash.h"
#include "FEN.h"
#include "Evaluador.h"
#include "Config.h"
#include "TableroDebug.h"

namespace UCI {

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ class ConsolaUCI
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Constructor
ConsolaUCI::ConsolaUCI(void) {
    // La salida por pantalla
    salida = new SalidaUCI;

    // Lista de comandos estandar
    listaComandos.push_back(new ComandoUCI("quit", this, &ConsolaUCI::cmdQuit));
    listaComandos.push_back(new ComandoUCI("ucinewgame", this, &ConsolaUCI::cmdUciNewGame));
    listaComandos.push_back(new ComandoUCI("uci", this, &ConsolaUCI::cmdUci));
    listaComandos.push_back(new ComandoUCI("isready", this, &ConsolaUCI::cmdIsReady));
    listaComandos.push_back(new ComandoUCI("position", this, &ConsolaUCI::cmdPosition));
    listaComandos.push_back(new ComandoUCI("go", this, &ConsolaUCI::cmdGo, false, true));
    listaComandos.push_back(new ComandoUCI("stop", this, &ConsolaUCI::cmdStop));
    listaComandos.push_back(new ComandoUCI("quit", this, &ConsolaUCI::cmdQuit));
    listaComandos.push_back(new ComandoUCI("setoption", this, &ConsolaUCI::cmdSetOption));

    // Lista de comandos extra
#if defined(DEBUG_MOSTRAR)
    listaComandos.push_back(new ComandoUCI("tablero", this, &ConsolaUCI::cmdTablero, true));
    listaComandos.push_back(new ComandoUCI("bitboards", this, &ConsolaUCI::cmdBitboards, true));
    listaComandos.push_back(new ComandoUCI("listamovs", this, &ConsolaUCI::cmdListaMovs, true));
    listaComandos.push_back(new ComandoUCI("eval", this, &ConsolaUCI::cmdEval, true));
    listaComandos.push_back(new ComandoUCI("move", this, &ConsolaUCI::cmdMove, true));
    listaComandos.push_back(new ComandoUCI("position_test", this, &ConsolaUCI::cmdPositionTest, true));
    listaComandos.push_back(new ComandoUCI("see", this, &ConsolaUCI::cmdSEE, true));
#endif
#if defined(USAR_PERFT)
    listaComandos.push_back(new ComandoUCI("perft", this, &ConsolaUCI::cmdPerft, true));
    listaComandos.push_back(new ComandoUCI("divide", this, &ConsolaUCI::cmdDivide, true));
#endif

    // Lista de opciones estandar
    listaOpciones.push_back(new OpcionUCISpin("Hash", this, &ConsolaUCI::optHash, 1024, 1, Config::sizeHashMegas));
    listaOpciones.push_back(new OpcionUCIButton("Clear Hash", this, &ConsolaUCI::optClearHash));
    listaOpciones.push_back(new OpcionUCIString("UCI_EngineAbout", this, &ConsolaUCI::optUciEngineAbout, Config::engineAbout));

    // Lista de opciones del motor
    listaOpciones.push_back(new OpcionUCICheck("Resumen Busqueda", this, &ConsolaUCI::optResumenBusqueda, Config::mostrarResumenBusqueda));
}

// Destructor
ConsolaUCI::~ConsolaUCI(void) {
    // La salida por pantalla
    delete salida;

    // Eliminamos la lista de comandos
    for(vector<ComandoUCI*>::iterator it = listaComandos.begin(); it != listaComandos.end(); it++) 
        delete (*it);

    // Eliminamos la lista de opciones
    for(vector<OpcionUCI*>::iterator it = listaOpciones.begin(); it != listaOpciones.end(); it++)
        delete (*it);
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ PROCESO INPUT
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Procesa la input recibida
void ConsolaUCI::ProcesarInput(void) {
    string::size_type longitudNombre;
    string argumentos = "";
    string comando;

    // Obtenemos el nombre del comando
    comando = PrimerPalabra(comandoRecibido);

    // Se pasa entre la lista de comandos viendo si se lo encuentra
    for(vector<ComandoUCI*>::iterator it = listaComandos.begin(); it != listaComandos.end(); it++) {  
        if(comando == (*it)->GetNombre()) {
            // Se encontr el comando, ahora se separa lo dems de la linea como argumento
            longitudNombre = (*it)->GetNombre().length();
            if(comandoRecibido.length() > comando.length())
                argumentos = comandoRecibido.substr(longitudNombre+1, string::npos);
            
            // Se llama a la funcin que maneja el comando
            (*it)->Ejecutar(argumentos);
            
            // Se manda la sear de que ya se proces el comando
            if(!(*it)->MandaSignal())
                InputProcesada();

            return;
        }
    }
    
    // Comando no encontrado, se manda la seal igual
    InputProcesada();
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ LISTA DE COMANDOS ESTANDAR
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Para una explicacin de los comandos UCI:
// http://download.shredderchess.com/div/uci.zip
// http://en.wikipedia.org/wiki/Universal_Chess_Interface


// Comando uci
void ConsolaUCI::cmdUci(string &args) {
    Salida() << "id name HeavyChess Alpha [" << VERSION << "]" << endl;
	Salida() << "id author Federico Corigliano" << endl;
    for(vector<OpcionUCI*>::iterator it = listaOpciones.begin(); it != listaOpciones.end(); it++)
        (*it)->Mostrar();
    Salida() << "uciok" << endl;
}

// Comando isready
void ConsolaUCI::cmdIsReady(string &args) {
	Salida() << "readyok" << endl;
}

// Comando position
void ConsolaUCI::cmdPosition(string &args) {
    vector<string> tokens = Tokenizar(args, " ");
    string::size_type i;
    string::size_type posMoves = string::npos;

    if(tokens.size() == 0)
        return;

    // Posicion inicial
    if(tokens[0] == "startpos") {
        GetMotor()->GetTablero().EstablecerPosicionInicial();        
        if(tokens.size() > 2)
            posMoves = 2;
    }
    // FEN
    else if(tokens[0] == "fen") {
        string sfen = "";

        for(i = 1; i < tokens.size() && tokens[i] != "moves"; i++) {
            sfen += tokens[i];
            sfen += " ";
        }

        if(!GetMotor()->GetTablero().LeerFEN(FEN(sfen))) {
            Salida() << "Cadena FEN corrupta" << endl;    
        }

        if(i+1 < tokens.size() && tokens[i] == "moves")
            posMoves = i+1;
    }

    // Movimientos extra
    if(posMoves != string::npos) {
        Movimiento m;
        for(i = posMoves; i < tokens.size(); i++) {
            m = Notador::ObtenerMovimiento(tokens[i], GetMotor()->GetTablero());
            if(!m.EsNulo()) {
                GetMotor()->GetTablero().Hacer(m);
                if(!GetMotor()->GetTablero().PosicionLegal()) {
                    GetMotor()->GetTablero().Deshacer();
                    Salida() << "Movimiento erroneo: " << tokens[i] << endl;
                }
            }
            else {
                Salida() << "Movimiento erroneo: " << tokens[i] << endl;
            }
        }
    }
}

// Comando go
void ConsolaUCI::cmdGo(string &args) {
    vector<string> tokens = Tokenizar(args, " ");

    if(tokens.size() == 0)
        return;
    
    // Detectamos el control de tiempo
    ControlTiempo &controlTiempo = GetMotor()->GetBusqueda().GetControlTiempo();
    controlTiempo.Limpiar();
    vector<string>::iterator it = tokens.begin();
    while(it != tokens.end()) {
        if(*it == "wtime" && (it+1) != tokens.end()) {
            controlTiempo.SetDispBlancas(StringAEntero(*(it+1)));
            it++;
        }
        else if(*it == "btime" && (it+1) != tokens.end()) {
            controlTiempo.SetDispNegras(StringAEntero(*(it+1)));
            it++;
        }
        else if(*it == "winc" && (it+1) != tokens.end()) {
            controlTiempo.SetIncBlancas(StringAEntero(*(it+1)));
            it++;
        }
        else if(*it == "binc" && (it+1) != tokens.end()) {
            controlTiempo.SetIncNegras(StringAEntero(*(it+1)));
            it++;
        }    
        else if(*it == "btime" && (it+1) != tokens.end()) {
            controlTiempo.SetDispNegras(StringAEntero(*(it+1)));
            it++;
        }    
        else if(*it == "movestogo" && (it+1) != tokens.end()) {
            controlTiempo.SetMovesToGo(StringAEntero(*(it+1)));
            it++;
        }
        else if(*it == "movetime" && (it+1) != tokens.end()) {
            controlTiempo.SetTiempoFijo(StringAEntero(*(it+1)));
            it++;
        }
        else if(*it == "depth" && (it+1) != tokens.end()) {
            controlTiempo.SetProfundidad(StringAEntero(*(it+1)));
            it++;
        }
        else if(*it == "nodes" && (it+1) != tokens.end()) {
            // Implementar nodes
            it++;
        }
        else if(*it == "mate" && (it+1) != tokens.end()) {
            // Implementar mate
            it++;
        }
        else if(*it == "infinite") {
            controlTiempo.SetInfinito();
        }
        else if(*it == "ponder") {
            // Implementar ponder
        }
        else if(*it == "searchmoves" && (it+1) != tokens.end()) {
            // Implementar searchmoves
            it++;
        }
        it++;
    }
    
    // Llamamos a la busqueda
    detenerMotor = false;
	InputProcesada();
    Movimiento m = GetMotor()->GetBusqueda().CalcularMejorMovimiento();
    Salida() << "bestmove " << m << endl;
}

// Comando stop
void ConsolaUCI::cmdStop(string &args) {
    DetenerMotor();
    GetMotor()->GetBusqueda().PararBusqueda();
}

// Comando quit
void ConsolaUCI::cmdQuit(string &args) {
	CerrarMotor();
    DetenerMotor();
    GetMotor()->GetBusqueda().PararBusqueda();
}

// Comando UCI2: ucinewgame
void ConsolaUCI::cmdUciNewGame(string &args) {
    GetMotor()->GetTablero().EstablecerPosicionInicial();
    GetMotor()->GetTablaHash().Limpiar();
}

// Comando set option name
void ConsolaUCI::cmdSetOption(string &args) {
    string::size_type longitudOpcion;
    string valor = "";
    string opcion;

    // Obtenemos el nombre del comando
    opcion = args.substr(args.find("name")+5, string::npos);
    opcion = PrimerPalabra(opcion);

    // Se pasa entre la lista de comandos viendo si se lo encuentra
    for(vector<OpcionUCI*>::iterator it = listaOpciones.begin(); it != listaOpciones.end(); it++) {
        if(opcion == (*it)->GetNombre()) {
            longitudOpcion = (*it)->GetNombre().length();
            // Se encontr la opcin, ahora se obtiene el valor
            string::size_type temp = args.find("value");
            string valor = args.substr(temp+6, string::npos);

            // Establecemos el nuevo valor
            (*it)->SetValor(valor);            
           
            return;
        }
    }
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ LISTA DE COMANDOS EXTRA
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Muestra el tablero
void ConsolaUCI::cmdTablero(string &args) {
#if defined(DEBUG_MOSTRAR)
    GetMotor()->GetTableroDebug().Mostrar();
#endif
}

// Muestra los bitboards
void ConsolaUCI::cmdBitboards(string &args) {
#if defined(DEBUG_MOSTRAR)
    GetMotor()->GetTableroDebug().MostrarBitboards();
#endif
}

// Calculo de perft
void ConsolaUCI::cmdPerft(string &args) {
#if defined (USAR_PERFT)
    int profundidad = StringAEntero(args);
    if(profundidad > 0)
        GetMotor()->GetBusqueda().CalcularPerft(profundidad);
#endif
}

// Calculo de PerftDivide
void ConsolaUCI::cmdDivide(string &args) {
#if defined (USAR_PERFT)
    int profundidad = StringAEntero(args);
    if(profundidad > 0)
        GetMotor()->GetBusqueda().CalcularPerftDivide(profundidad);
#endif
}

// Muestra la lista de movimientos disponibles
void ConsolaUCI::cmdListaMovs(string &args) {
#if defined(DEBUG_MOSTRAR)
    Salida() << GetMotor()->GetTablero().ListaMovimientosLegales() << endl;
    Salida() << "Total Movimientos Legales: " << GetMotor()->GetTablero().TotalMovimientosLegales() << endl;
#endif
}

// Muestra la evaluacin esttica
void ConsolaUCI::cmdEval(string &args) {
#if defined(DEBUG_MOSTRAR)
    Salida() << "Eval: " << GetMotor()->GetEval().Evaluar(GetMotor()->GetTablero()) << endl;
#endif
}

// Realiza un movimiento sobre el tablero
void ConsolaUCI::cmdMove(string &args) {
#if defined(DEBUG_MOSTRAR)
    Movimiento m = Notador::ObtenerMovimiento(args, GetMotor()->GetTablero());
    if(!m.EsNulo()) {
        GetMotor()->GetTablero().Hacer(m);
        if(!GetMotor()->GetTablero().PosicionLegal()) {
            GetMotor()->GetTablero().Deshacer();
            Salida() << "Movimiento Ilegal: " << args << endl;
        }
    }
    else {
        Salida() << "Movimiento Ilegal: " << args << endl;
    }
#endif
}

// Carga una posicion fen para testeo generador movimientos
void ConsolaUCI::cmdPositionTest(string &args) {
#if defined(DEBUG_MOSTRAR)
    int num = StringAEntero(args);
    switch(num) {
        case 1:
            GetMotor()->GetTablero().LeerFEN(FEN("r3k2r/p1ppqpb1/bn2pnp1/3PN3/1p2P3/2N2Q1p/PPPBBPPP/R3K2R w KQkq - 0 1"));
            GetMotor()->GetTableroDebug().Mostrar();
            Salida() << "perft(6) = 8031647685" << endl;
            break;
        case 2:
            GetMotor()->GetTablero().LeerFEN(FEN("8/3K4/2p5/p2b2r1/5k2/8/8/1q6 b - - 0 1"));
            GetMotor()->GetTableroDebug().Mostrar();
            Salida() << "perft(2) = 279" << endl;
            break;
        case 3:
            GetMotor()->GetTablero().LeerFEN(FEN("8/7p/p5pb/4k3/P1pPn3/8/P5PP/1rB2RK1 b - d3 0 1"));
            GetMotor()->GetTableroDebug().Mostrar();
            Salida() << "perft(6) = 38633283" << endl;
            break;
        case 4:
            GetMotor()->GetTablero().LeerFEN(FEN("rnbqkb1r/ppppp1pp/7n/4Pp2/8/8/PPPP1PPP/RNBQKBNR w KQkq f6 0 1"));
            GetMotor()->GetTableroDebug().Mostrar();
            Salida() << "perft(5) = 11139762" << endl;
            break;
        case 5:
            GetMotor()->GetTablero().LeerFEN(FEN("8/2p5/3p4/KP5r/1R3p1k/8/4P1P1/8 w - - 0 1"));
            GetMotor()->GetTableroDebug().Mostrar();
            Salida() << "perft(7) = 178633661" << endl;
            break;
        case 6:
            GetMotor()->GetTablero().LeerFEN(FEN("8/k7/3p4/p2P1p2/P2P1P2/8/8/K7 w - - 0 1"));
            GetMotor()->GetTableroDebug().Mostrar();
            Salida() << "Fine #70" << endl;
            break;
    };
#endif
}

// Para testear el SEE
void ConsolaUCI::cmdSEE(std::string &args) {
#if defined(DEBUG_MOSTRAR)
    Movimiento m = Notador::ObtenerMovimiento(args, GetMotor()->GetTablero());
    Salida() << "SEE[" << Notador::ObtenerCoordenadas(m.Origen()) << "][" << Notador::ObtenerCoordenadas(m.Destino()) << "] = ";
    Salida() << GetMotor()->GetTablero().SEE(m) << endl;
#endif
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ LISTA DE OPCIONES ESTANDAR
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Para establecer el tamao de la tabla hash
void ConsolaUCI::optHash(string &valor) {
    int size = StringAEntero(valor);
    if(size)
        GetMotor()->GetTablaHash().Alojar(size);
}

// Para limpia el contenido de la tabla hash
void ConsolaUCI::optClearHash(string &valor) {
    GetMotor()->GetTablaHash().Limpiar();
}

// Muestra info sobre el motor
void ConsolaUCI::optUciEngineAbout(string &valor) {
    Config::engineAbout = valor;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ LISTA DE OPCIONES EXTRA
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Habilita o deshabilita el resumen de estadsticas de bsqueda
void ConsolaUCI::optResumenBusqueda(string &valor) {
    if(valor == "true")
        GetMotor()->GetBusqueda().HabilitarResumen();
    else if(valor == "false")
        GetMotor()->GetBusqueda().DeshabilitarResumen();
}

}; // namespace UCI
