//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ EntradaHash.h
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#pragma once

#include "Hashkeys.h"

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ MACROS
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ Representacion de la info hash en un integer
//+ 
//+  3130292827262524232221201918171615141312111009080706050403020100
//+ +----------------------------------------------------------------+
//+ +SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSLLLLPPPPPPPPPPPPPPAAEEEEEE      +
//+ +----------------------------------------------------------------+
//+
//+ E = Edad (3 bits) Offset = 3
//+ A = Amenaza (1 bit) Offset = 6
//+ P = Profundidad (7 bits) Offset = 7
//+ L = Tipo Limite (2 bits) Offset = 14
//+ S = Puntaje (16 bits) Offset = 16
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


// Para extraer la informacin
#define HASH_PUNTAJE(x)             (x >> 16)
#define HASH_TIPO_LIMITE(x)         ((x >> 14) & 3)
#define HASH_PROFUNDIDAD(x)         ((x >> 7) & 127)
#define HASH_AMENAZA(x)             ((x >> 6) & 1)
#define HASH_EDAD(x)                ((x >> 3) & 3)

// Para guardar informacin
#define SET_HASH_PUNTAJE(x)         (x << 16)
#define SET_HASH_TIPO_LIMITE(x)     ((x & 3) << 14)
#define SET_HASH_PROFUNDIDAD(x)     ((x & 127) << 7)
#define SET_HASH_AMENAZA(x)         ((x & 1) << 6)
#define SET_HASH_EDAD(x)            ((x & 3) << 3)

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ TIPOS ENUMERADOS
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

enum HashTipoLimite {
    HashSinInfo,
    HashAlpha,
    HashBeta,
    HashExacto,
    HashInfoSecundaria
};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ class EntradaHash
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class EntradaHash {
public:
    // Setters
    void SetMovimiento(Movimiento m);
    void SetHashkey(Hashkey h);
    void SetFlags(int profundidad, short int puntaje, HashTipoLimite tipoLimite, bool amenaza, int edad);

    // Getters
    Hashkey GetHashkey(void) const;
    Movimiento GetMovimiento(void) const;
    int GetPuntaje(void) const;
    HashTipoLimite GetTipoLimite(void) const;
    bool GetAmenazaMate(void) const;
    int GetProfundidad(void) const;
    int GetEdad(void) const;

private:

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ DATOS
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    Hashkey hashkey;        // La clave hash de la posicin
    Movimiento movimiento;  // El mejor movimiento encontrado para la posicin
    int flags;              // Los flags: edad, amenaza de mate, profundidad, puntaje, tipo de lmite
};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ FUNCIONES INLINE
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Establece el movimiento de la entrada
inline void EntradaHash::SetMovimiento(Movimiento m) {
    movimiento = m;
}

// Establece la hashkey de la posicin
inline void EntradaHash::SetHashkey(Hashkey h) {
    hashkey = h;
}

// Establece los flags de la posicin
inline void EntradaHash::SetFlags(int profundidad, short int puntaje, HashTipoLimite tipoLimite, bool amenaza, int edad = 0) {
    flags = SET_HASH_PROFUNDIDAD(profundidad)
          | SET_HASH_PUNTAJE(puntaje)
          | SET_HASH_TIPO_LIMITE(tipoLimite)
          | SET_HASH_EDAD(edad)
          | SET_HASH_AMENAZA(amenaza);
}

// Devuelve la hashkey en la entrada
inline Hashkey EntradaHash::GetHashkey(void) const {
    return hashkey;
}

// Devuelve el movimiento en la entrada
inline Movimiento EntradaHash::GetMovimiento(void) const {
    return movimiento;
}

// Devuelve el puntaje de la entrada
inline int EntradaHash::GetPuntaje(void) const {
    return HASH_PUNTAJE(flags);
}

// Devuelve el tipo de lmite de la entrada
inline HashTipoLimite EntradaHash::GetTipoLimite(void) const {
    return static_cast<HashTipoLimite>(HASH_TIPO_LIMITE(flags));
}

// Devuelve si hay amenaza de mate en la posicin
inline bool EntradaHash::GetAmenazaMate(void) const {
    return HASH_AMENAZA(flags);
}

// Devuelve la profunidad de la entrada
inline int EntradaHash::GetProfundidad(void) const {
    return HASH_PROFUNDIDAD(flags);
}

// Devuelve la edad de la entrada
inline int EntradaHash::GetEdad(void) const {
    return HASH_EDAD(flags);
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ class EntradaHashDoble
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class EntradaHashDoble {
public:
    // Se que esto est mal hecho, pero por ahora me sirve
    EntradaHash* GetEntradaDP(void) { return &entradaDP; }
    EntradaHash* GetEntradaAR(void) { return &entradaAR; }
private:
    EntradaHash entradaDP;
    EntradaHash entradaAR;
};
