//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ FEN.cpp
//+
//+ Para representar una posicin de ajedrez en notacin FEN (Forsyth-Edwards Notation)
//+ Para ms informacin: http://en.wikipedia.org/wiki/FEN
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "stdafx.h"
#include "FEN.h"
#include "Notador.h"

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ FUNCIONES AUXILIARES
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ class FEN
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Constructor
FEN::FEN(void) : cadenaFEN(""), posValida(false) {}

// Constructor
FEN::FEN(std::string fen) : cadenaFEN(fen) {
    posValida = false;
    if(Procesar())
        posValida = true;
}

// Devuelve el ply
int FEN::Ply(void) const {
    return ply;
}

// Devuelve la pieza en la casilla c
Pieza FEN::GetPieza(Casilla c) const {
    assert(CasillaValida(c));
    return casillas[c];
}

// Devuelve el estado del tablero
EstadoTablero FEN::GetEstado(void) const {
    return estado;
}

// Establece la pieza en la casilla c
void FEN::Set(Casilla c, Pieza p) {
    assert(CasillaValida(c));
    assert(PiezaValida(p));
    casillas[c] = p;
}

// Devuelve true si el FEN es vlido
bool FEN::EsValida(void) const {
    return posValida;
}

// Limpia los datos del objeto
void FEN::LimpiarDatos(void) {
    for(int i = 0; i < 64; i++)
        casillas[i] = Vacio;

    estado.Limpiar(); 
    ply = 1;
    posValida = false;
}

// Lee una cadena FEN alojada en un string y la procesa
void FEN::SetCadena(string fen) {
    cadenaFEN = fen;
    if(Procesar())
        posValida = true;
}

// Carga los datos desde la cadena FEN leida
bool FEN::Procesar(void) {
    // Se limpia la representacin interna del tablero y se
    // tokeniza el string
    LimpiarDatos();
    vector<string> tokens = Tokenizar(cadenaFEN, " ");

    // Se procesa el string sabiendo en que fase estamos segn el token en el que se est presente
    int fase = 1;
    for(vector<string>::iterator it = tokens.begin(); it != tokens.end(); it++, fase++) {
        switch(fase) {
            // Fase 1: las piezas en el tablero
            case 1: {
                Casilla c = A8;
                for(unsigned int i = 0; i < (*it).length(); i++, c++) {
                    switch((*it)[i]) {
                        case 'K': Set(c,ReyB); break;
                        case 'Q': Set(c,DamaB); break;
                        case 'R': Set(c,TorreB); break;
                        case 'B': Set(c,AlfilB); break;
                        case 'N': Set(c,CaballoB); break;
                        case 'P': Set(c,PeonB); break;
                        case 'k': Set(c,ReyN); break;
                        case 'q': Set(c,DamaN); break;
                        case 'r': Set(c,TorreN); break;
                        case 'b': Set(c,AlfilN); break;
                        case 'n': Set(c,CaballoN); break;
                        case 'p': Set(c,PeonN); break;
                        case '1':
                        case '2':
                        case '3':
                        case '4':
                        case '5':
                        case '6':
                        case '7':
                        case '8': c+=((*it)[i])-'1'; break;
                        case '/': c-=17;
                    }
                }
                break;
            }
            // Fase 2: Lado que mueve        
            case 2:
                estado.SetLadoMueve((*it == "w") ? Blancas : Negras);
                break;
            // Fase 3: Derechos de enroque
            case 3:
                if((*it).find("K")!=string::npos) estado.GetInfoEnroque().SetBlancasPuedenEnrocarCorto();
                if((*it).find("k")!=string::npos) estado.GetInfoEnroque().SetNegrasPuedenEnrocarCorto();
                if((*it).find("Q")!=string::npos) estado.GetInfoEnroque().SetBlancasPuedenEnrocarLargo();
                if((*it).find("q")!=string::npos) estado.GetInfoEnroque().SetNegrasPuedenEnrocarLargo();
                break;
            // Fase 4: Casilla de captura al paso
            case 4:
                if(*it != "-")
                    estado.SetCasillaAlPaso(Notador::ObtenerCasilla(*it));
                break;
            // Fase 5: nmero de jugadas para la regla de los 50 movimientos
            case 5:
                estado.SetCincuenta(StringAEntero(*it));
                break;
            // Fase 6: ply, nmero de medios-movimientos desde el inicio de la partida
            case 6:
                ply = StringAEntero(*it);
                break;
        }
    }

    // No ha habido errores, la cadena FEN es vlida
    return true;
}
