//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ ListaMovimientos.cpp
//+
//+ Representa una lista de movimientos posibles sobre el tablero
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "stdafx.h"
#include "ListaMovimientos.h"
#include "Notador.h"
#include "Tablero.h"
#include "Busqueda.h"
#include "Estadisticas.h"

using namespace Estadisticas;

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ class ListaMovimientos
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Establece el tipo de lista
void ListaMovimientos::SetTipoLista(TipoLista tipoLista) {
    tipo = tipoLista;
    movActual = cantidad = 0;
    switch(tipo) {
        case COMPLETA:
            siguienteFase = PROBAR_HASH;
            break;
        case CAPTURAS:
            siguienteFase = GENERAR_CAPTURAS;
            break;
        case PERFT:
            siguienteFase = PASAR_RESTO_MOVS;
            tablero->GenerarCapturas(*this);
            tablero->GenerarNoCapturas(*this);
            break;
        case EVASIONES:
            siguienteFase = PROBAR_HASH;
            tablero->GenerarEvasiones(*this);
            evasionesGeneradas++;
            break;
        case COMPLETA_RAIZ: {
            siguienteFase = PROBAR_HASH;
            tablero->GenerarCapturas(*this);
            tablero->GenerarNoCapturas(*this);
            
            // Sacamos los movimientos ilegales de entrada 
            // utilizando una copia del tablero
            Tablero tableroCopia(*tablero);
            for(int i = 0; i < cantidad; i++) {
                tableroCopia.Hacer(lista[i]);
                if(!tableroCopia.PosicionLegal()) {
                    if(cantidad > 1) {
                        lista[i] = lista[cantidad-1];
                        cantidad--;
                        i--; // Para no saltarse el movimiento que colocamos ahora donde el estaba el ilegal
                    }
                    else {
                        cantidad = 0;
                    }
                }
                tableroCopia.Deshacer();
            }

            capturasGeneradas++;
            nocapturasGeneradas++;
            break;
        }
    }
}

// Va al principio de la lista de movimientos
void ListaMovimientos::Reiniciar(void) {
    movActual = 0;
    switch(tipo) {
        case COMPLETA:
            // Ac se generan de nuevo los movimientos
            cantidad = 0;
            siguienteFase = PROBAR_HASH;
            break;
        case PERFT:
            siguienteFase = PASAR_RESTO_MOVS;
            break;
        case COMPLETA_RAIZ:
            siguienteFase = PROBAR_HASH;
            break;
        case EVASIONES:
            siguienteFase = PROBAR_HASH;
            break;
        case CAPTURAS:
            siguienteFase = PASAR_RESTO_MOVS;
            break;
    }
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ ORDENADO DE MOVIMIENTOS
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Se utiliza el siguiente orden:
// 1) Movimiento de la hash
// 2) Capturas que ganan o no pierden material (SEE >= 0)
// 3) Movimientos killer (1 y 2)
// 4) No Capturas, ordenadas por History Heuristic
// 5) Capturas que pierden material
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Devuelve el siguiente movimiento en la lista (para listas tipos COMPLETA y PERFT)
Movimiento ListaMovimientos::GetSiguienteCompleta(int ply) {
    Movimiento m;
    puntajeActual = 0;

    while(true) {
        switch(siguienteFase) {
            // El movimiento de la tabla hash
            case PROBAR_HASH:
                siguienteFase = GENERAR_CAPTURAS;
                m = arbol->GetMovHash(ply);
                if(!m.EsNulo())
                    return m;
                break;

            // Generacin de capturas
            case GENERAR_CAPTURAS:                
                siguienteFase = ORDENAR_CAPTURAS;
                tablero->GenerarCapturas(*this);
                capturasGeneradas++;
                break;

            // Ordenado capturas
            case ORDENAR_CAPTURAS:
                siguienteFase = PROBAR_CAPTURAS_GANADORAS;

                // Asignamos el puntaje para cada captura
                for(int i = 0; i < cantidad; i++) {                    
                    // Si una de las capturas es el movimiento de la hash, lo sacamos
                    if(lista[i] == arbol->GetMovHash(ply)) {
                        puntajes[i] = PUNTAJE_EVITAR_MOVIMIENTO;
                    }
                    else {
                        // Puntaje segn el SEE
                        puntajes[i] = tablero->SEE(lista[i]);
                    }
                }
                
                // Se ordenan los movimientos
                Ordenar();
                break;

            // Capturas que ganan material o no pierden material (SEE >= 0)
            case PROBAR_CAPTURAS_GANADORAS:
                siguienteFase = PROBAR_KILLER1;
                if(cantidad > 0 && puntajes[0] >= 0) {
                    m = lista[0];
                    puntajeActual = puntajes[0];
                    puntajes[0] = PUNTAJE_EVITAR_MOVIMIENTO;
                    return m;
                }
                break;

            // El Killer1
            case PROBAR_KILLER1:
                siguienteFase = PROBAR_KILLER2;
                m = arbol->GetKiller1(ply);
                if(tablero->EsMovimientoValido(m) && m != arbol->GetMovHash(ply))
                    return m;
                break;

            // El Killer2
            case PROBAR_KILLER2:
                siguienteFase = GENERAR_NOCAPTURAS;
                m = arbol->GetKiller2(ply);
                if(tablero->EsMovimientoValido(m) && m != arbol->GetMovHash(ply))
                    return m;
                break;
              
            // Generacin del resto de los movimientos
            case GENERAR_NOCAPTURAS:
                siguienteFase = ORDENAR_MOVIMIENTOS;
                tablero->GenerarNoCapturas(*this);
                nocapturasGeneradas++;
                break;

            // Ordenado del resto de los movimientos
            case ORDENAR_MOVIMIENTOS:
                siguienteFase = PASAR_RESTO_MOVS;

                for(int i = 0; i < cantidad; i++) {
                    // Si unos de los movimientos es el de hash o algun killer lo sacamos
                    if(lista[i] == arbol->GetMovHash(ply) || lista[i] == arbol->GetKiller1(ply) || lista[i] == arbol->GetKiller2(ply)) {
                        puntajes[i] = PUNTAJE_EVITAR_MOVIMIENTO;
                    }
                    else if(!lista[i].Captura() && !lista[i].Promocion()) {
                        // Puntaje segn heurstica historia para las no capturas
                        // Si es una captura ya tiene un puntaje asignado
                        puntajes[i] = arbol->GetHH(lista[i].GetPieza(), lista[i].Destino());
                    }
                }

                // Ordenamos los movimientos
                Ordenar();
                break;
            
            // Probamos con el resto de los movimientos
            case PASAR_RESTO_MOVS:               
                if(movActual >= cantidad || puntajes[movActual] == PUNTAJE_EVITAR_MOVIMIENTO) {
                    siguienteFase = FIN_MOVS;
                    break;
                }

                puntajeActual = puntajes[movActual];
                return lista[movActual++];
                break;

            // Se terminaron los movimientos
            case FIN_MOVS:                
                m.SetNulo();
                return m;
                break;
        }
    }
}

// Devuelve la siguiente captura en la lista (para lista tipo CAPTURAS)
Movimiento ListaMovimientos::GetSiguienteCaptura(int ply) {
    Movimiento m;

    while(true) {
        switch(siguienteFase) {
            // Generacin de capturas
            case GENERAR_CAPTURAS:                
                siguienteFase = ORDENAR_CAPTURAS;
                tablero->GenerarCapturas(*this);
                capturasGeneradas++;
                break;

            // Ordenado de las capturas
            case ORDENAR_CAPTURAS:
                siguienteFase = PASAR_RESTO_MOVS;

                // Asignamos el puntaje para cada captura
                for(int i = 0; i < cantidad; i++) {
                    // Puntaje segn el SEE
                    puntajes[i] = tablero->SEE(lista[i]);
                }
                
                // Se ordenan los movimientos
                Ordenar();
                break;
            
            // Probamos con el resto de los movimientos
            case PASAR_RESTO_MOVS:               
                if(movActual >= cantidad) {
                    siguienteFase = FIN_MOVS;
                    break;
                }

                puntajeActual = puntajes[movActual];
                return lista[movActual++];
                break;

            // Se terminaron los movimientos
            case FIN_MOVS:                
                m.SetNulo();
                return m;
                break;
        }
    }
}

// Devuelve la siguiente evasin al jaque en la lista (para lista tipo EVASIONES)
Movimiento ListaMovimientos::GetSiguienteEvasion(int ply) {
    Movimiento m;
    puntajeActual = 0;

    while(true) {
        switch(siguienteFase) {
            // El movimiento de la tabla hash
            case PROBAR_HASH:
                siguienteFase = ORDENAR_MOVIMIENTOS;
                m = arbol->GetMovHash(ply);
                if(!m.EsNulo())
                    return m;
                break;

            // Ordenado de movimientos
            case ORDENAR_MOVIMIENTOS:                
                siguienteFase = PASAR_RESTO_MOVS;

                // Asignamos puntajes a los movimientos:
                // Si es una captura, segn el SEE
                // Si es una no-captura, segn la historia
                for(int i = 0; i < cantidad; i++) {
                    // Si una de las capturas es el movimiento de la hash, lo sacamos
                    if(lista[i] == arbol->GetMovHash(ply)) {
                        puntajes[i] = PUNTAJE_EVITAR_MOVIMIENTO;
                    }
                    else if(lista[i].Captura() || lista[i].Promocion()) {
                        // Puntaje segn el SEE
                        puntajes[i] = tablero->SEE(lista[i]);
                    }
                    else {
                        // Puntaje segn heurstica historia
                        puntajes[i] = arbol->GetHH(lista[i].GetPieza(), lista[i].Destino());                        
                    }
                }

                // Se ordenan los movimientos
                Ordenar();
                break;
            
            // Probamos con el resto de los movimientos
            case PASAR_RESTO_MOVS:               
                if(movActual >= cantidad || puntajes[movActual] == PUNTAJE_EVITAR_MOVIMIENTO) {
                    siguienteFase = FIN_MOVS;
                    break;
                }

                puntajeActual = puntajes[movActual];
                return lista[movActual++];
                break;

            // Se terminaron los movimientos
            case FIN_MOVS:                
                m.SetNulo();
                return m;
                break;
        }
    }
}

// Devuelve el siguiente movimiento (para lista tipo COMPLETA_RAIZ)
Movimiento ListaMovimientos::GetSiguienteRaiz(int ply) {
    Movimiento m;
    puntajeActual = 0;

    while(true) {
        switch(siguienteFase) {
            // El movimiento de la tabla hash
            case PROBAR_HASH:
                siguienteFase = ORDENAR_MOVIMIENTOS;
                m = arbol->GetMovHash(ply);
                if(!m.EsNulo())
                    return m;
                break;

            // Ordenado de movimientos
            case ORDENAR_MOVIMIENTOS:                
                siguienteFase = PASAR_RESTO_MOVS;

                // Asignamos puntajes a los movimientos segn el nmero de nodos del subarbol
                // generado por cada movimiento en el ply anterior                
                
                // Primero buscamos el movimiento de la hash y le damos el puntaje del
                // primer movimiento visto. Esto es as porque la lista tiene un orden ac
                // que no es el que se cumpli pues el movimiento de la hash siempre va primero
                if(!arbol->GetMovHash(ply).EsNulo()) {
                    for(int i = 0; i < cantidad; i++) {
                        if(arbol->GetMovHash(ply) == lista[i])
                            puntajes[i] = static_cast<int>(arbol->GetNumeroNodosRaiz(0));
                    }

                    // Les damos al resto de los movimientos su puntaje
                    for(int i = 0, j = 1; i < cantidad; i++) {
                        if(arbol->GetMovHash(ply) == lista[i])
                            continue;
                        puntajes[i] = static_cast<int>(arbol->GetNumeroNodosRaiz(j));
                        j++;
                    }
                }
                else {
                    // No hubo movimiento de la hash en el ply, entonces simplemente
                    // se ordena por los nodos
                    for(int i = 0; i < cantidad; i++)
                        puntajes[i] = static_cast<int>(arbol->GetNumeroNodosRaiz(i));
                }


                // Se ordenan los movimientos
                Ordenar();
                break;
            
            // Probamos con el resto de los movimientos
            case PASAR_RESTO_MOVS:               
                if(movActual >= cantidad) {
                    siguienteFase = FIN_MOVS;
                    break;
                }

                // Saltamos ac al movimiento de la hash porque ya se lo vi
                if(arbol->GetMovHash(ply) == lista[movActual]) {
                    movActual++;
                    continue;
                }

                puntajeActual = puntajes[movActual];
                return lista[movActual++];
                break;

            // Se terminaron los movimientos
            case FIN_MOVS:                
                m.SetNulo();
                return m;
                break;
        }
    }
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ QUICKSORT
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Ordena los movimientos en la lista
void ListaMovimientos::Ordenar(void) {
    int temp, i, j;
    Movimiento temp2;

    // Ordenado por insercin
    // Prob con quicksort recursivo y burbuja tambin, pero la insercin
    // dio los mejores resultados
    for(i = 0; i < cantidad; i++) {
        temp = puntajes[i];
        temp2 = lista[i];
        j = i - 1;
        while( (puntajes[j] < temp) && (j >= 0) ) {
            lista[j+1] = lista[j];
            puntajes[j+1] = puntajes[j];
            j--;
        }
        puntajes[j+1] = temp;
        lista[j+1] = temp2;
    }
}
