//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ Tablero.cpp
//+
//+ Para representar un tablero de ajedrez en el programa junto con las
//+ funciones necesarias para utilizarlo
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "stdafx.h"
#include "Tablero.h"
#include "Notador.h"
#include "ListaMovimientos.h"
#include "FEN.h"
#include "Estadisticas.h"
#include "TableroDebug.h"

using namespace Bitboards;
using namespace Hashkeys;

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ class Tablero
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Constructor
Tablero::Tablero(void) {
    EstablecerPosicionInicial();
    tableroDebug = NULL;
}

// Constructor Copia
Tablero::Tablero(const Tablero &tablero) {
    for(int i = 0; i < 64; i++) casillas[i] = tablero.casillas[i];    
    estado = tablero.estado;
    material[Blancas] = tablero.material[Blancas];
    material[Negras] = tablero.material[Negras];
    totalValorPiezas[Blancas] = tablero.totalValorPiezas[Blancas];
    totalValorPiezas[Negras] = tablero.totalValorPiezas[Negras];
    totalPeones[Blancas] = tablero.totalPeones[Blancas];
    totalPeones[Negras] = tablero.totalPeones[Negras];
    posReyB = tablero.posReyB;
    posReyN = tablero.posReyN;
    hashkey = tablero.hashkey;

    historial = tablero.historial;

    piezasBlancas = tablero.piezasBlancas;
    piezasNegras = tablero.piezasNegras;
    for(int i = 0; i < MAX_PIEZAS; i++) piezas[i] = tablero.piezas[i];
    rotado90 = tablero.rotado90;
    rotado45 = tablero.rotado45;
    rotado315 = tablero.rotado315;

    tableroDebug = tablero.tableroDebug;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ EL ANALIZADOR DEL TABLERO
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void Tablero::SetTableroDebug(const TableroDebug *td) {
    tableroDebug = td;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ ESTABLECIMIENTO DE POSICIONES
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Establece la posicion inicial de una partida de ajedrez
void Tablero::EstablecerPosicionInicial(void) {
    LeerFEN(FEN("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"));
}

// Limpia el tablero, dejando todo en blanco
void Tablero::Limpiar(void) {
    for(Casilla i = A1; i <= H8; i++)
        SetPieza(i,Vacio);
    estado.Limpiar();
    historial.Limpiar();
    material[Blancas] = material[Negras] = 0;
    totalPeones[Blancas] = totalPeones[Negras] = 0;
    totalValorPiezas[Blancas] = totalValorPiezas[Negras] = 0;
    hashkey = 0;
}

// Regenera los bitboards, material y otras estructuras que se actualizan automaticamente
void Tablero::RegenerarEstructuras(void) {
    piezasBlancas = piezasNegras = 0;
    for(int i = 0; i < MAX_PIEZAS; i++)
        piezas[i] = 0;

    for(Casilla c = A1; c <= H8; c++) {
        Pieza pieza = GetPieza(c);
        if(pieza != Vacio) {
            AgregarBit(piezas[pieza],c);
            if(ColorPieza(pieza) == Blancas) {
                AgregarBit(piezasBlancas,c);
                AgregarMaterial(pieza);
            }
            else {
                AgregarBit(piezasNegras,c);
                AgregarMaterial(pieza);
            }
        }
    }

    rotado90 = RotarBitboard90(Ocupadas());
    rotado45 = RotarBitboard45(Ocupadas());
    rotado315 = RotarBitboard315(Ocupadas());
    hashkey = GenerarHashkey();
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ LECTURA DE POSICIONES FEN
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Configura el tablero segn una cadena FEN
bool Tablero::LeerFEN(FEN &fen) {
    Limpiar();
    if(fen.EsValida()) {
        for(Casilla i = A1; i <= H8; i++) {
            SetPieza(i,fen.GetPieza(i));
            if(GetPieza(i) == ReyB) SetPosicionReyB(i);
            if(GetPieza(i) == ReyN) SetPosicionReyN(i);
        }
        estado = fen.GetEstado();
        RegenerarEstructuras();
        return true;
    }
    return false;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ GENERACION DE MOVIMIENTOS
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Genera capturas y promociones
void Tablero::GenerarCapturas(ListaMovimientos &listaMovs) const {
    Bitboard casillasVacias = Vacias();
    Bitboard casillasOcupadas = Ocupadas();
    Bitboard destinos;
    Bitboard origen;
    Casilla i, d;

    if(estado.BlancasMueven()) {
        // Capturas Caballos
        origen = CaballosB();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesCaballo(i) & piezasNegras;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarCaptura(i, d, CaballoB, GetPieza(d));
            }
        }

        // Capturas Reyes
        origen = ReyesB();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesRey(i) & piezasNegras;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarCaptura(i, d, ReyB, GetPieza(d));
            }
        }

        // Capturas Torres-Damas
        origen = TorresB() | DamasB();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesTorre(i) & piezasNegras;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarCaptura(i, d, GetPieza(i), GetPieza(d));
            }
        }

        // Capturas Alfiles-Damas
        origen = AlfilesB() | DamasB();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesAlfil(i) & piezasNegras;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarCaptura(i, d, GetPieza(i), GetPieza(d));
            }
        }

        // Promociones Peones Normales
        destinos = ((PeonesB() & MaskFila[Fila7])<<8) & casillasVacias;
        while(destinos) {
            d = MSB(destinos);
            SacarBit(destinos,d);
            listaMovs.AgregarPromocionBlanca(d-8,d,DamaB,GetPieza(d));
            listaMovs.AgregarPromocionBlanca(d-8,d,CaballoB,GetPieza(d));
            listaMovs.AgregarPromocionBlanca(d-8,d,TorreB,GetPieza(d));
            listaMovs.AgregarPromocionBlanca(d-8,d,AlfilB,GetPieza(d));
        }

        // Capturas Al Paso
        if(estado.GetCasillaAlPaso() != Invalida) {
            Casilla ap = estado.GetCasillaAlPaso();
            if(Columna(ap) < ColH && GetPieza(ap-7) == PeonB)
                listaMovs.AgregarAlPasoBlanco(ap-7,ap);
            if(Columna(ap) > ColA && GetPieza(ap-9) == PeonB)
                listaMovs.AgregarAlPasoBlanco(ap-9,ap);
        }

        // Capturas Hacia la Izquierda
        destinos = ((PeonesB() & ~MaskCol[ColA])<<7) & piezasNegras;
        while(destinos) {
            d = MSB(destinos);
            SacarBit(destinos,d);
            if(Fila(d) == Fila8) {
                listaMovs.AgregarPromocionBlanca(d-7,d,DamaB,GetPieza(d));
                listaMovs.AgregarPromocionBlanca(d-7,d,CaballoB,GetPieza(d));
                listaMovs.AgregarPromocionBlanca(d-7,d,TorreB,GetPieza(d));
                listaMovs.AgregarPromocionBlanca(d-7,d,AlfilB,GetPieza(d));
            }
            else
                listaMovs.AgregarCaptura(d-7,d,PeonB,GetPieza(d));
        }

        // Capturas Hacia la Derecha
        destinos = ((PeonesB() & ~MaskCol[ColH])<<9) & piezasNegras;
        while(destinos) {
            d = MSB(destinos);
            SacarBit(destinos,d);
            if(Fila(d) == Fila8) {
                listaMovs.AgregarPromocionBlanca(d-9,d,DamaB,GetPieza(d));
                listaMovs.AgregarPromocionBlanca(d-9,d,CaballoB,GetPieza(d));
                listaMovs.AgregarPromocionBlanca(d-9,d,TorreB,GetPieza(d));
                listaMovs.AgregarPromocionBlanca(d-9,d,AlfilB,GetPieza(d));
            }
            else
                listaMovs.AgregarCaptura(d-9,d,PeonB,GetPieza(d));
        }
    }
    else {
        // Capturas Caballos
        origen = CaballosN();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesCaballo(i) & piezasBlancas;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarCaptura(i, d, CaballoN, GetPieza(d));
            }
        }

        // Capturas Reyes
        origen = ReyesN();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesRey(i) & piezasBlancas;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarCaptura(i, d, ReyN, GetPieza(d));
            }
        }

        // Capturas Torres-Damas
        origen = TorresN() | DamasN();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesTorre(i) & piezasBlancas;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarCaptura(i, d, GetPieza(i), GetPieza(d));
            }
        }

        // Capturas Alfiles-Damas
        origen = AlfilesN() | DamasN();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesAlfil(i) & piezasBlancas;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarCaptura(i, d, GetPieza(i), GetPieza(d));
            }
        }

        // Promociones Peones Normales
        destinos = ((PeonesN() & MaskFila[Fila2])>>8) & casillasVacias;
        while(destinos) {
            d = MSB(destinos);
            SacarBit(destinos,d);
            listaMovs.AgregarPromocionNegra(d+8,d,DamaN,GetPieza(d));
            listaMovs.AgregarPromocionNegra(d+8,d,CaballoN,GetPieza(d));
            listaMovs.AgregarPromocionNegra(d+8,d,TorreN,GetPieza(d));
            listaMovs.AgregarPromocionNegra(d+8,d,AlfilN,GetPieza(d));
        }

        // Capturas Al Paso
        if(estado.GetCasillaAlPaso() != Invalida) {
            Casilla ap = estado.GetCasillaAlPaso();
            if(Columna(ap) > ColA && GetPieza(ap+7) == PeonN)
                listaMovs.AgregarAlPasoNegro(ap+7,ap);
            if(Columna(ap) < ColH && GetPieza(ap+9) == PeonN)
                listaMovs.AgregarAlPasoNegro(ap+9,ap);
        }

        // Capturas Hacia la Izquierda
        destinos = ((PeonesN() & ~MaskCol[ColA])>>9) & piezasBlancas;
        while(destinos) {
            d = MSB(destinos);
            SacarBit(destinos,d);
            if(Fila(d) == Fila1) {
                listaMovs.AgregarPromocionNegra(d+9,d,DamaN,GetPieza(d));
                listaMovs.AgregarPromocionNegra(d+9,d,CaballoN,GetPieza(d));
                listaMovs.AgregarPromocionNegra(d+9,d,TorreN,GetPieza(d));
                listaMovs.AgregarPromocionNegra(d+9,d,AlfilN,GetPieza(d));
            }
            else
                listaMovs.AgregarCaptura(d+9,d,PeonN,GetPieza(d));
        }

        // Capturas Hacia la Derecha
        destinos = ((PeonesN() & ~MaskCol[ColH])>>7) & piezasBlancas;
        while(destinos) {
            d = MSB(destinos);
            SacarBit(destinos,d);
            if(Fila(d) == Fila1) {
                listaMovs.AgregarPromocionNegra(d+7,d,DamaN,GetPieza(d));
                listaMovs.AgregarPromocionNegra(d+7,d,CaballoN,GetPieza(d));
                listaMovs.AgregarPromocionNegra(d+7,d,TorreN,GetPieza(d));
                listaMovs.AgregarPromocionNegra(d+7,d,AlfilN,GetPieza(d));
            }
            else
                listaMovs.AgregarCaptura(d+7,d,PeonN,GetPieza(d));
        }
    }
}

// Genera no capturas: movimientos simples y enroques
void Tablero::GenerarNoCapturas(ListaMovimientos &listaMovs) const {
    Bitboard casillasVacias = Vacias();
    Bitboard casillasOcupadas = Ocupadas();
    Bitboard destinos, destinos2;
    Bitboard origen;
    Casilla i, d;
    
    if(estado.BlancasMueven()) {
        // Movimientos Caballos
        origen = CaballosB();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesCaballo(i) & casillasVacias;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarSimple(i, d, CaballoB);
            }
        }

        // Movimientos Reyes
        origen = ReyesB();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesRey(i) & casillasVacias;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarSimple(i, d, ReyB);
            }
        }

        // Movimientos Torres-Damas
        origen = TorresB() | DamasB();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesTorre(i) & casillasVacias;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarSimple(i, d, GetPieza(i));
            }
        }

        // Movimientos Alfiles-Damas
        origen = AlfilesB() | DamasB();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesAlfil(i) & casillasVacias;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarSimple(i, d, GetPieza(i));
            }
        }

        // Enroques
        if(estado.GetInfoEnroque().BlancasPuedenEnrocarCorto() && !(casillasOcupadas & (Mask[F1]|Mask[G1])) && !EsAtacado(E1,Negras) && !EsAtacado(F1,Negras))
            listaMovs.AgregarECB();
        if(estado.GetInfoEnroque().BlancasPuedenEnrocarLargo() && !(casillasOcupadas & (Mask[D1]|Mask[C1]|Mask[B1])) && !EsAtacado(D1,Negras) && !EsAtacado(E1,Negras))
            listaMovs.AgregarELB();
        
        // Peones
        destinos = ((PeonesB() & ~MaskFila[Fila7])<<8) & casillasVacias;
        // Peones Doble Movimiento en Casilla Inicial
        destinos2 = ((destinos & MaskFila[Fila3])<<8) & casillasVacias;

        while(destinos) {
            d = MSB(destinos);
            SacarBit(destinos,d);
            listaMovs.AgregarSimple(d-8,d,PeonB);
        }

        while(destinos2) {
            d = MSB(destinos2);
            SacarBit(destinos2,d);
            listaMovs.AgregarSimple(d-16,d,PeonB);
        }
    }
    else {
        // Movimientos Caballos
        origen = CaballosN();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesCaballo(i) & casillasVacias;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarSimple(i, d, CaballoN);
            }
        }

        // Movimientos Reyes
        origen = ReyesN();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesRey(i) & casillasVacias;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarSimple(i, d, ReyN);
            }
        }

        // Movimientos Torres-Damas
        origen = TorresN() | DamasN();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesTorre(i) & casillasVacias;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarSimple(i, d, GetPieza(i));
            }
        }

        // Movimientos Alfiles-Damas
        origen = AlfilesN() | DamasN();
        while(origen) {
            i = MSB(origen);
            SacarBit(origen,i);
            destinos = AtaquesAlfil(i) & casillasVacias;
            while(destinos) {
                d = MSB(destinos);
                SacarBit(destinos, d);
                listaMovs.AgregarSimple(i, d, GetPieza(i));
            }
        }

        // Enroques
        if(estado.GetInfoEnroque().NegrasPuedenEnrocarCorto() && !(casillasOcupadas & (Mask[F8]|Mask[G8])) && !EsAtacado(F8,Blancas) && !EsAtacado(E8,Blancas))
            listaMovs.AgregarECN();
        if(estado.GetInfoEnroque().NegrasPuedenEnrocarLargo() && !(casillasOcupadas & (Mask[D8]|Mask[C8]|Mask[B8])) && !EsAtacado(D8,Blancas) && !EsAtacado(E8,Blancas))
            listaMovs.AgregarELN();

        // Peones
        destinos = ((PeonesN() & ~MaskFila[Fila2])>>8) & casillasVacias;
        // Peones Doble Movimiento en Casilla Inicial
        destinos2 = ((destinos & MaskFila[Fila6])>>8) & casillasVacias;

        while(destinos) {
            d = MSB(destinos);
            SacarBit(destinos,d);
            listaMovs.AgregarSimple(d+8,d,PeonN);
        }

        while(destinos2) {
            d = MSB(destinos2);
            SacarBit(destinos2,d);
            listaMovs.AgregarSimple(d+16,d,PeonN);
        }
    }
}

// Genera evasiones legales al jaque
void Tablero::GenerarEvasiones(ListaMovimientos &listaMovs) const {
    Bitboard piezasJaque;
    Bitboard destinos;
    Bitboard destinos2;
    Bitboard obstruir;
    Bitboard origen;
    Casilla casillaAtaque;
    Casilla casillaAtaque2;
    Casilla destino;
    Casilla posicionRey;
    Casilla i;
    Direccion direccion1 = SIN_DIRECCION;
    Direccion direccion2 = SIN_DIRECCION;

    // Evasiones de las piezas blancas al jaque negro
    if(GetEstado().BlancasMueven()) {
		//----------------------------------------------------
		// Deteccin de atacantes
		//----------------------------------------------------

        // Deteccin de las piezas que dan jaque al rey, como mximo son dos
        posicionRey = GetPosicionReyB();
        piezasJaque = AtaquesCasilla(posicionRey) & piezasNegras;

        // La primer pieza
        casillaAtaque = MSB(piezasJaque);
        SacarBit(piezasJaque, casillaAtaque);

        // La direccin del primer ataque
        if(GetPieza(casillaAtaque) != PeonN)
            direccion1 = Direcciones[casillaAtaque][posicionRey];

        // La direccin del segundo ataque (si existe)
        if(piezasJaque) {
            casillaAtaque2 = MSB(piezasJaque);
            if(GetPieza(casillaAtaque2) != PeonN)
                direccion2 = Direcciones[casillaAtaque2][posicionRey];
        }

		//----------------------------------------------------
		// Escapes del rey
		//----------------------------------------------------
        
        destinos = AtaquesRey(posicionRey) & ~piezasBlancas;

        while(destinos) {
            destino = MSB(destinos);
            SacarBit(destinos, destino);
            if(!EsAtacado(destino, Negras) && Direcciones[posicionRey][destino] != direccion1 && Direcciones[posicionRey][destino] != direccion2)
                listaMovs.AgregarCaptura(posicionRey, destino, ReyB, GetPieza(destino));
        }

		//----------------------------------------------------
		// Obstrucciones al jaque: solo si hay un atacante
		//----------------------------------------------------

        if(!piezasJaque) {
            // Las casillas a donde una pieza se puede mover para obstruir el jaque
            obstruir = Mask[casillaAtaque] | MaskEntre[casillaAtaque][posicionRey];

		    //----------------------------------------------------
		    // Obstrucciones Caballos y piezas deslizantes
		    //----------------------------------------------------

            // Obstrucciones caballos
            origen = CaballosB();
            while(origen) {
                i = MSB(origen);
                SacarBit(origen, i);
                if(!ClavadaReyBlanco(i)) {
                    destinos = AtaquesCaballo(i) & obstruir;
                    while(destinos) {
                        destino = MSB(destinos);
                        SacarBit(destinos, destino);
                        listaMovs.AgregarCaptura(i, destino, CaballoB, GetPieza(destino));
                    }
                }
            }

            // Obstrucciones alfiles-damas
            origen = AlfilesB() | DamasB();
            while(origen) {
                i = MSB(origen);
                SacarBit(origen, i);
                if(!ClavadaReyBlanco(i)) {
                    destinos = AtaquesAlfil(i) & obstruir;
                    while(destinos) {
                        destino = MSB(destinos);
                        SacarBit(destinos, destino);
                        listaMovs.AgregarCaptura(i, destino, GetPieza(i), GetPieza(destino));
                    }
                }
            }

            // Obstrucciones torres-damas
            origen = TorresB() | DamasB();
            while(origen) {
                i = MSB(origen);
                SacarBit(origen, i);
                if(!ClavadaReyBlanco(i)) {
                    destinos = AtaquesTorre(i) & obstruir;
                    while(destinos) {
                        destino = MSB(destinos);
                        SacarBit(destinos, destino);
                        listaMovs.AgregarCaptura(i, destino, GetPieza(i), GetPieza(destino));
                    }
                }
            }

		    //----------------------------------------------------
		    // Obstrucciones peones
		    //----------------------------------------------------

            // Capturas al paso que capturen la pieza que da jaque (necesariamente debe ser un
            // pen y el jaque lo da al dar un salto doble)
            if(FilaAbajo(estado.GetCasillaAlPaso()) == casillaAtaque) {
                Casilla ap = estado.GetCasillaAlPaso();
                if(Columna(ap) < ColH && GetPieza(ap-7) == PeonB && !ClavadaReyBlanco(ap-7))
                    listaMovs.AgregarAlPasoBlanco(ap-7,ap);
                if(Columna(ap) > ColA && GetPieza(ap-9) == PeonB && !ClavadaReyBlanco(ap-9))
                    listaMovs.AgregarAlPasoBlanco(ap-9,ap);
            }

            // Capturas Hacia la Izquierda que obstruyan
            destinos = ((PeonesB() & ~MaskCol[ColA])<<7) & piezasNegras & obstruir;
            while(destinos) {
                destino = MSB(destinos);
                SacarBit(destinos,destino);
                if(!ClavadaReyBlanco(destino-7)) {
                    if(Fila(destino) == Fila8) {
                        listaMovs.AgregarPromocionBlanca(destino-7,destino,DamaB,GetPieza(destino));
                        listaMovs.AgregarPromocionBlanca(destino-7,destino,CaballoB,GetPieza(destino));
                        listaMovs.AgregarPromocionBlanca(destino-7,destino,TorreB,GetPieza(destino));
                        listaMovs.AgregarPromocionBlanca(destino-7,destino,AlfilB,GetPieza(destino));
                    }
                    else
                        listaMovs.AgregarCaptura(destino-7,destino,PeonB,GetPieza(destino));
                }
            }

            // Capturas Hacia la Derecha que obstruyan
            destinos = ((PeonesB() & ~MaskCol[ColH])<<9) & piezasNegras & obstruir;
            while(destinos) {
                destino = MSB(destinos);
                SacarBit(destinos,destino);
                if(!ClavadaReyBlanco(destino-9)) {
                    if(Fila(destino) == Fila8) {
                        listaMovs.AgregarPromocionBlanca(destino-9,destino,DamaB,GetPieza(destino));
                        listaMovs.AgregarPromocionBlanca(destino-9,destino,CaballoB,GetPieza(destino));
                        listaMovs.AgregarPromocionBlanca(destino-9,destino,TorreB,GetPieza(destino));
                        listaMovs.AgregarPromocionBlanca(destino-9,destino,AlfilB,GetPieza(destino));
                    }
                    else
                        listaMovs.AgregarCaptura(destino-9,destino,PeonB,GetPieza(destino));
                }
            }            

            // Peones
            destinos = (PeonesB()<<8) & Vacias();
            // Peones Doble Movimiento en Casilla Inicial
            destinos2 = ((destinos & MaskFila[Fila3])<<8) & Vacias() & obstruir;
            destinos &= obstruir;

            // Una casilla
            while(destinos) {
                destino = MSB(destinos);
                SacarBit(destinos,destino);
                if(!ClavadaReyBlanco(destino-8)) {
                    if(Fila(destino) == Fila8) {
                        // Promociones que obstruyen
                        listaMovs.AgregarPromocionBlanca(destino-8,destino,DamaB,GetPieza(destino));
                        listaMovs.AgregarPromocionBlanca(destino-8,destino,CaballoB,GetPieza(destino));
                        listaMovs.AgregarPromocionBlanca(destino-8,destino,TorreB,GetPieza(destino));
                        listaMovs.AgregarPromocionBlanca(destino-8,destino,AlfilB,GetPieza(destino));
                    }
                    else {
                        // Movimientos comunes
                        listaMovs.AgregarSimple(destino-8,destino,PeonB);
                    }
                }                
            }

            // Dos casillas
            while(destinos2) {
                destino = MSB(destinos2);
                SacarBit(destinos2,destino);
                if(!ClavadaReyBlanco(destino-16)) {
                    listaMovs.AgregarSimple(destino-16,destino,PeonB);
                }
            }
        }
    }
    // Evasiones de las piezas negras al jaque blanco
    else {
		//----------------------------------------------------
		// Deteccin de atacantes
		//----------------------------------------------------

        // Deteccin de las piezas que dan jaque al rey, como mximo son dos
        posicionRey = GetPosicionReyN();
        piezasJaque = AtaquesCasilla(posicionRey) & piezasBlancas;

        // La primer pieza
        casillaAtaque = MSB(piezasJaque);
        SacarBit(piezasJaque, casillaAtaque);

        // La direccin del primer ataque
        if(GetPieza(casillaAtaque) != PeonB)
            direccion1 = Direcciones[casillaAtaque][posicionRey];

        // La direccin del segundo ataque (si existe)
        if(piezasJaque) {
            casillaAtaque2 = MSB(piezasJaque);
            if(GetPieza(casillaAtaque2) != PeonB)
                direccion2 = Direcciones[casillaAtaque2][posicionRey];
        }

		//----------------------------------------------------
		// Escapes del rey
		//----------------------------------------------------
        
        destinos = AtaquesRey(posicionRey) & ~piezasNegras;

        while(destinos) {
            destino = MSB(destinos);
            SacarBit(destinos, destino);
            if(!EsAtacado(destino, Blancas) && Direcciones[posicionRey][destino] != direccion1 && Direcciones[posicionRey][destino] != direccion2)
                listaMovs.AgregarCaptura(posicionRey, destino, ReyN, GetPieza(destino));
        }

		//----------------------------------------------------
		// Obstrucciones al jaque: solo si hay un atacante
		//----------------------------------------------------

        if(!piezasJaque) {
            // Las casillas a donde una pieza se puede mover para obstruir el jaque
            obstruir = Mask[casillaAtaque] | MaskEntre[casillaAtaque][posicionRey];

		    //----------------------------------------------------
		    // Obstrucciones Caballos y piezas deslizantes
		    //----------------------------------------------------

            // Obstrucciones caballos
            origen = CaballosN();
            while(origen) {
                i = MSB(origen);
                SacarBit(origen, i);
                if(!ClavadaReyNegro(i)) {
                    destinos = AtaquesCaballo(i) & obstruir;
                    while(destinos) {
                        destino = MSB(destinos);
                        SacarBit(destinos, destino);
                        listaMovs.AgregarCaptura(i, destino, CaballoN, GetPieza(destino));
                    }
                }
            }

            // Obstrucciones alfiles-damas
            origen = AlfilesN() | DamasN();
            while(origen) {
                i = MSB(origen);
                SacarBit(origen, i);
                if(!ClavadaReyNegro(i)) {
                    destinos = AtaquesAlfil(i) & obstruir;
                    while(destinos) {
                        destino = MSB(destinos);
                        SacarBit(destinos, destino);
                        listaMovs.AgregarCaptura(i, destino, GetPieza(i), GetPieza(destino));
                    }
                }
            }

            // Obstrucciones torres-damas
            origen = TorresN() | DamasN();
            while(origen) {
                i = MSB(origen);
                SacarBit(origen, i);
                if(!ClavadaReyNegro(i)) {
                    destinos = AtaquesTorre(i) & obstruir;
                    while(destinos) {
                        destino = MSB(destinos);
                        SacarBit(destinos, destino);
                        listaMovs.AgregarCaptura(i, destino, GetPieza(i), GetPieza(destino));
                    }
                }
            }

		    //----------------------------------------------------
		    // Obstrucciones peones
		    //----------------------------------------------------

            // Capturas al paso que capturen la pieza que da jaque (necesariamente debe ser un
            // pen y el jaque lo da al dar un salto doble)
            if(FilaArriba(estado.GetCasillaAlPaso()) == casillaAtaque) {
                Casilla ap = estado.GetCasillaAlPaso();
                if(Columna(ap) < ColH && GetPieza(ap+7) == PeonN && !ClavadaReyNegro(ap+7))
                    listaMovs.AgregarAlPasoNegro(ap+7,ap);
                if(Columna(ap) > ColA && GetPieza(ap+9) == PeonN && !ClavadaReyNegro(ap+9))
                    listaMovs.AgregarAlPasoNegro(ap+9,ap);
            }

            // Capturas Hacia la Izquierda que obstruyan
            destinos = ((PeonesN() & ~MaskCol[ColA])>>9) & piezasBlancas & obstruir;
            while(destinos) {
                destino = MSB(destinos);
                SacarBit(destinos,destino);
                if(!ClavadaReyNegro(destino+9)) {
                    if(Fila(destino) == Fila1) {
                        // Promociones con captura
                        listaMovs.AgregarPromocionNegra(destino+9,destino,DamaN,GetPieza(destino));
                        listaMovs.AgregarPromocionNegra(destino+9,destino,CaballoN,GetPieza(destino));
                        listaMovs.AgregarPromocionNegra(destino+9,destino,TorreN,GetPieza(destino));
                        listaMovs.AgregarPromocionNegra(destino+9,destino,AlfilN,GetPieza(destino));
                    }
                    else {
                        // Capturas
                        listaMovs.AgregarCaptura(destino+9,destino,PeonN,GetPieza(destino));
                    }
                }
            }

            // Capturas Hacia la Derecha que obstruyan
            destinos = ((PeonesN() & ~MaskCol[ColH])>>7) & piezasBlancas & obstruir;
            while(destinos) {
                destino = MSB(destinos);
                SacarBit(destinos,destino);
                if(!ClavadaReyNegro(destino+7)) {
                    if(Fila(destino) == Fila1) {
                        // Promociones
                        listaMovs.AgregarPromocionNegra(destino+7,destino,DamaN,GetPieza(destino));
                        listaMovs.AgregarPromocionNegra(destino+7,destino,CaballoN,GetPieza(destino));
                        listaMovs.AgregarPromocionNegra(destino+7,destino,TorreN,GetPieza(destino));
                        listaMovs.AgregarPromocionNegra(destino+7,destino,AlfilN,GetPieza(destino));
                    }
                    else {
                        // Capturas
                        listaMovs.AgregarCaptura(destino+7,destino,PeonN,GetPieza(destino));
                    }
                }
            }            

            // Peones
            destinos = (PeonesN()>>8) & Vacias();
            // Peones Doble Movimiento en Casilla Inicial
            destinos2 = ((destinos & MaskFila[Fila6])>>8) & Vacias() & obstruir;
            destinos &= obstruir;

            // Una casilla
            while(destinos) {
                destino = MSB(destinos);
                SacarBit(destinos,destino);
                if(!ClavadaReyNegro(destino+8)) {
                    if(Fila(destino) == Fila1) {
                        // Promociones que obstruyen
                        listaMovs.AgregarPromocionNegra(destino+8,destino,DamaN,GetPieza(destino));
                        listaMovs.AgregarPromocionNegra(destino+8,destino,CaballoN,GetPieza(destino));
                        listaMovs.AgregarPromocionNegra(destino+8,destino,TorreN,GetPieza(destino));
                        listaMovs.AgregarPromocionNegra(destino+8,destino,AlfilN,GetPieza(destino));
                    }
                    else {
                        // Movimientos comunes
                        listaMovs.AgregarSimple(destino+8,destino,PeonN);
                    }
                }                
            }

            // Dos casillas
            while(destinos2) {
                destino = MSB(destinos2);
                SacarBit(destinos2,destino);
                if(!ClavadaReyNegro(destino+16)) {
                    listaMovs.AgregarSimple(destino+16,destino,PeonN);
                }
            }
        }
    }
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ HACER Y DESHACER
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Realiza un movimiento en el tablero
void Tablero::Hacer(Movimiento m) {
    Bitboard M;
    Casilla origen = m.Origen();
    Casilla destino = m.Destino();
    Pieza captura = m.Captura();
    Pieza pieza = m.GetPieza();
    
    // Chequeamos que el movimiento sea vlido
    assert(!m.EsNulo());

    // Agregamos el movimiento al historial
    historial.Agregar(m,hashkey,estado);
    historial.AumentarPly();
    
    // Cambio de lado
    ActualizarHashLado(hashkey);
    estado.CambiarLado();

    // Actualizacin casilla al paso
    ActualizarHashAlPaso(hashkey,estado.GetCasillaAlPaso());
    estado.SetCasillaAlPaso(Invalida);
            
    // Regla 50 movimientos
    estado.IncrementarCincuenta();

Principio:

    // Mascara para actualizar bitboards piezas
    M = Mask[origen] | Mask[destino];

    // Actualizacion bitboards rotados
    XOR(rotado45, Mask45[origen] | Mask45[destino]);
    XOR(rotado315, Mask315[origen] | Mask315[destino]);
    XOR(rotado90, Mask90[origen] | Mask90[destino]);

    // Borramos la pieza de origen
    SetPieza(origen,Vacio);
    ActualizarHash(hashkey, pieza, origen);

    // Definimos la operacin segn la pieza que mueve
    switch(pieza) {
        case ReyB:
            XOR(piezasBlancas, M);
            SetPieza(destino, ReyB);
            SetPosicionReyB(destino);
            ActualizarHashEnroque(hashkey, estado.GetInfoEnroque()); // Si movemos el rey se cae el enroque
            estado.GetInfoEnroque().DesactivarEnroquesBlancos();
            ActualizarHashEnroque(hashkey, estado.GetInfoEnroque());
            ActualizarHash(hashkey, ReyB, destino);
            if(m.EsECB()) {
                origen = H1; // Movimiento de la torre
                destino = F1;
                pieza = TorreB;
                goto Principio;
            }
            else if(m.EsELB()) {
                origen = A1; // Movimiento de la torre
                destino = D1;
                pieza = TorreB;
                goto Principio;
            }         
            break;
        case ReyN:
            XOR(piezasNegras, M);
            SetPieza(destino, ReyN);
            SetPosicionReyN(destino);
            ActualizarHashEnroque(hashkey, estado.GetInfoEnroque()); // Si movemos el rey se cae el enroque
            estado.GetInfoEnroque().DesactivarEnroquesNegros();
            ActualizarHashEnroque(hashkey, estado.GetInfoEnroque());
            ActualizarHash(hashkey, ReyN, destino);
            if(m.EsECN()) {
                origen = H8; // Movimiento de la torre
                destino = F8;
                pieza = TorreN;
                goto Principio;
            }
            else if(m.EsELN()) {
                origen = A8; // Movimiento de la torre
                destino = D8;
                pieza = TorreN;
                goto Principio;
            }
            break;
        case PeonB:
            XOR(piezasBlancas, M);
            XOR(PeonesB(), M);
            SetPieza(destino, PeonB);
            estado.ReiniciarReglaCincuenta();
            ActualizarHash(hashkey, PeonB, destino);
            if(m.EsAP()) {
                SacarBit45(rotado45,FilaAbajo(destino)); // Captura al paso
                SacarBit315(rotado315,FilaAbajo(destino));
                SacarBit90(rotado90,FilaAbajo(destino));
                SacarBit(PeonesN(), FilaAbajo(destino));
                SacarBit(piezasNegras, FilaAbajo(destino));
                SetPieza(FilaAbajo(destino), Vacio);
                ActualizarHash(hashkey, PeonN, FilaAbajo(destino));
                QuitarMaterial(PeonN);
                captura = Vacio;
            }
            else if(m.Promocion()) {
                SacarBit(PeonesB(), destino);
                QuitarMaterial(PeonB);
                ActualizarHash(hashkey, PeonB, destino);
                switch(m.Promocion()) { // Promocion
                    case DamaB:
                        AgregarBit(DamasB(), destino);
                        SetPieza(destino, DamaB);
                        AgregarMaterial(DamaB);
                        ActualizarHash(hashkey, DamaB, destino);
                        break;
                    case TorreB:
                        AgregarBit(TorresB(), destino);
                        SetPieza(destino, TorreB);
                        AgregarMaterial(TorreB);
                        ActualizarHash(hashkey, TorreB, destino);
                        break;
                    case AlfilB:
                        AgregarBit(AlfilesB(), destino);
                        SetPieza(destino, AlfilB);
                        AgregarMaterial(AlfilB);
                        ActualizarHash(hashkey, AlfilB, destino);
                        break;
                    case CaballoB:
                        AgregarBit(CaballosB(), destino);
                        SetPieza(destino, CaballoB);
                        AgregarMaterial(CaballoB);
                        ActualizarHash(hashkey, CaballoB, destino);
                        break;
                }
            }
            else if((destino == DosFilasArriba(origen)) && MaskAlPaso[destino] & piezasNegras) {
                estado.SetCasillaAlPaso(FilaArriba(origen)); // Nueva casilla al paso
                ActualizarHashAlPaso(hashkey,estado.GetCasillaAlPaso());
            }
            break;
        case PeonN:
            XOR(piezasNegras, M);
            XOR(PeonesN(), M);
            SetPieza(destino, PeonN);
            estado.ReiniciarReglaCincuenta();
            ActualizarHash(hashkey, PeonN, destino);
            if(m.EsAP()) {
                SacarBit45(rotado45,FilaArriba(destino)); // Captura al paso
                SacarBit315(rotado315,FilaArriba(destino));
                SacarBit90(rotado90,FilaArriba(destino));
                SacarBit(PeonesB(), FilaArriba(destino));
                SacarBit(piezasBlancas, FilaArriba(destino));
                SetPieza(FilaArriba(destino), Vacio);
                ActualizarHash(hashkey, PeonB, FilaArriba(destino));
                QuitarMaterial(PeonB);
                captura = Vacio;
            }
            else if(m.Promocion()) {
                SacarBit(PeonesN(), destino);
                QuitarMaterial(PeonN);
                ActualizarHash(hashkey, PeonN, destino);
                switch(m.Promocion()) { // Promocion
                    case DamaN:
                        AgregarBit(DamasN(), destino);
                        SetPieza(destino, DamaN);
                        AgregarMaterial(DamaN);
                        ActualizarHash(hashkey, DamaN, destino);
                        break;
                    case TorreN:
                        AgregarBit(TorresN(), destino);
                        SetPieza(destino, TorreN);
                        AgregarMaterial(TorreN);
                        ActualizarHash(hashkey, TorreN, destino);
                        break;
                    case AlfilN:
                        AgregarBit(AlfilesN(), destino);
                        SetPieza(destino, AlfilN);
                        AgregarMaterial(AlfilN);
                        ActualizarHash(hashkey, AlfilN, destino);
                        break;
                    case CaballoN:
                        AgregarBit(CaballosN(), destino);
                        SetPieza(destino, CaballoN);
                        AgregarMaterial(CaballoN);
                        ActualizarHash(hashkey, CaballoN, destino);
                        break;
                }
            }
            else if((destino == DosFilasAbajo(origen)) && MaskAlPaso[destino] & piezasBlancas) {
                estado.SetCasillaAlPaso(FilaAbajo(origen)); // Nueva casilla al paso
                ActualizarHashAlPaso(hashkey,estado.GetCasillaAlPaso());
            }
            break;
        case CaballoB:
            XOR(piezasBlancas, M);
            XOR(CaballosB(), M);
            SetPieza(destino,CaballoB);
            ActualizarHash(hashkey, CaballoB, destino);
            break;
        case CaballoN:
            XOR(piezasNegras, M);
            XOR(CaballosN(), M);
            SetPieza(destino,CaballoN);
            ActualizarHash(hashkey, CaballoN, destino);
            break;
        case AlfilB:
            XOR(piezasBlancas, M);
            XOR(AlfilesB(), M);
            SetPieza(destino,AlfilB);
            ActualizarHash(hashkey, AlfilB, destino);
            break;
        case AlfilN:
            XOR(piezasNegras, M);
            XOR(AlfilesN(), M);
            SetPieza(destino,AlfilN);
            ActualizarHash(hashkey, AlfilN, destino);
            break;
        case TorreB:
            XOR(piezasBlancas, M);
            XOR(TorresB(), M);
            SetPieza(destino,TorreB);
            ActualizarHash(hashkey, TorreB, destino);
            if(estado.GetInfoEnroque().BlancasPuedenEnrocarCorto() && origen == H1) {
                ActualizarHashEnroque(hashkey, estado.GetInfoEnroque()); // Si movemos la torre se cae el enroque
                estado.GetInfoEnroque().DesactivarEnroqueCortoBlanco();
                ActualizarHashEnroque(hashkey, estado.GetInfoEnroque());
            }
            else if(estado.GetInfoEnroque().BlancasPuedenEnrocarLargo() && origen == A1) {
                ActualizarHashEnroque(hashkey, estado.GetInfoEnroque()); // Si movemos la torre se cae el enroque
                estado.GetInfoEnroque().DesactivarEnroqueLargoBlanco();
                ActualizarHashEnroque(hashkey, estado.GetInfoEnroque());
            }
            break;
        case TorreN:
            XOR(piezasNegras, M);
            XOR(TorresN(), M);
            SetPieza(destino,TorreN);
            ActualizarHash(hashkey, TorreN, destino);
            if(estado.GetInfoEnroque().NegrasPuedenEnrocarCorto() && origen == H8) {
                ActualizarHashEnroque(hashkey, estado.GetInfoEnroque()); // Si movemos la torre se cae el enroque
                estado.GetInfoEnroque().DesactivarEnroqueCortoNegro();
                ActualizarHashEnroque(hashkey, estado.GetInfoEnroque());
            }
            else if(estado.GetInfoEnroque().NegrasPuedenEnrocarLargo() && origen == A8) {
                ActualizarHashEnroque(hashkey, estado.GetInfoEnroque()); // Si movemos la torre se cae el enroque
                estado.GetInfoEnroque().DesactivarEnroqueLargoNegro();
                ActualizarHashEnroque(hashkey, estado.GetInfoEnroque());
            }
            break;
        case DamaB:
            XOR(piezasBlancas, M);
            XOR(DamasB(), M);
            SetPieza(destino,DamaB);
            ActualizarHash(hashkey, DamaB, destino);
            break;
        case DamaN:
            XOR(piezasNegras, M);
            XOR(DamasN(), M);
            SetPieza(destino,DamaN);
            ActualizarHash(hashkey, DamaN, destino);
            break;
    }

    // Si el movimiento es una captura
    if(captura) {
        AgregarBit45(rotado45,destino);
        AgregarBit315(rotado315,destino);
        AgregarBit90(rotado90,destino);
        QuitarMaterial(captura);
        ActualizarHash(hashkey, captura, destino);
        estado.ReiniciarReglaCincuenta();

        switch(captura) {
            case PeonB:
                SacarBit(PeonesB(), destino);
                SacarBit(piezasBlancas, destino);
                break;
            case PeonN:
                SacarBit(PeonesN(), destino);
                SacarBit(piezasNegras, destino);
                break;
            case CaballoB:
                SacarBit(CaballosB(), destino);
                SacarBit(piezasBlancas, destino);
                break;
            case CaballoN:
                SacarBit(CaballosN(), destino);
                SacarBit(piezasNegras, destino);
                break;
            case AlfilB:
                SacarBit(AlfilesB(), destino);
                SacarBit(piezasBlancas, destino);
                break;
            case AlfilN:
                SacarBit(AlfilesN(), destino);
                SacarBit(piezasNegras, destino);
                break;
            case TorreB:
                SacarBit(TorresB(), destino);
                SacarBit(piezasBlancas, destino);
                if(estado.GetInfoEnroque().BlancasPuedenEnrocarCorto() && destino == H1) {
                    ActualizarHashEnroque(hashkey, estado.GetInfoEnroque()); // Se captura la torre se cae el enroque
                    estado.GetInfoEnroque().DesactivarEnroqueCortoBlanco();
                    ActualizarHashEnroque(hashkey, estado.GetInfoEnroque());
                }
                else if(estado.GetInfoEnroque().BlancasPuedenEnrocarLargo() && destino == A1) {
                    ActualizarHashEnroque(hashkey, estado.GetInfoEnroque()); // Se captura la torre se cae el enroque
                    estado.GetInfoEnroque().DesactivarEnroqueLargoBlanco();
                    ActualizarHashEnroque(hashkey, estado.GetInfoEnroque());
                }                
                break;
            case TorreN:
                SacarBit(TorresN(), destino);
                SacarBit(piezasNegras, destino);
                if(estado.GetInfoEnroque().NegrasPuedenEnrocarCorto() && destino == H8) {
                    ActualizarHashEnroque(hashkey, estado.GetInfoEnroque()); // Se captura la torre se cae el enroque
                    estado.GetInfoEnroque().DesactivarEnroqueCortoNegro();
                    ActualizarHashEnroque(hashkey, estado.GetInfoEnroque());
                }
                else if(estado.GetInfoEnroque().NegrasPuedenEnrocarLargo() && destino == A8) {
                    ActualizarHashEnroque(hashkey, estado.GetInfoEnroque()); // Se captura la torre se cae el enroque
                    estado.GetInfoEnroque().DesactivarEnroqueLargoNegro();
                    ActualizarHashEnroque(hashkey, estado.GetInfoEnroque());
                }                
                break;
            case DamaB:
                SacarBit(DamasB(), destino);
                SacarBit(piezasBlancas, destino);
                break;
            case DamaN:
                SacarBit(DamasN(), destino);
                SacarBit(piezasNegras, destino);
                break;
        }
    }

#if defined(CHEQUEO_CONSISTENCIA)
    ChequearConsistencia();
#endif
}

// Deshace el ultimo movimiento realizado sobre el tablero
void Tablero::Deshacer(void) {
    Bitboard M;
    Casilla origen;
    Casilla destino;
    Pieza captura;
    Pieza pieza;

    // Obtenemos los datos del movimiento que hay que deshacer
    historial.ReducirPly();
    Movimiento m = historial.GetMovimiento();
    origen = m.Origen();
    destino = m.Destino();
    captura = m.Captura();
    pieza = m.GetPieza();
    
    // Reemplazamos alguna informacin
    estado = historial.GetEstadoTablero();
    hashkey = historial.GetHashkey();

Principio:
    // Mascara para bitboards piezas
    M = Mask[origen] | Mask[destino];

    // Actualizacion bitboards rotados
    XOR(rotado45, Mask45[origen] | Mask45[destino]);
    XOR(rotado315, Mask315[origen] | Mask315[destino]);
    XOR(rotado90, Mask90[origen] | Mask90[destino]);

    // Borramos la pieza de destino
    SetPieza(destino, Vacio);

    // Definimos la operacin segn la pieza que mueve
    switch(pieza) {
        case ReyB:
            XOR(piezasBlancas, M);
            SetPieza(origen, ReyB);
            SetPosicionReyB(origen);
            if(m.EsECB()) {
                origen = H1; // Movimiento de la torre
                destino = F1;
                pieza = TorreB;
                goto Principio;
            }
            else if(m.EsELB()) {
                origen = A1; // Movimiento de la torre
                destino = D1;
                pieza = TorreB;
                goto Principio;
            }
            break;
        case ReyN:
            XOR(piezasNegras, M);
            SetPieza(origen, ReyN);
            SetPosicionReyN(origen);
            if(m.EsECN()) {
                origen = H8; // Movimiento de la torre
                destino = F8;
                pieza = TorreN;
                goto Principio;
            }
            else if(m.EsELN()) {
                origen = A8; // Movimiento de la torre
                destino = D8;
                pieza = TorreN;
                goto Principio;
            }
            break;
        case PeonB:
            XOR(piezasBlancas, M);
            XOR(PeonesB(), M);
            SetPieza(origen, PeonB);
            if(m.EsAP()) {
                AgregarBit(PeonesN(), FilaAbajo(destino)); // Captura al paso
                AgregarBit(piezasNegras, FilaAbajo(destino));
                AgregarBit45(rotado45, FilaAbajo(destino));
                AgregarBit315(rotado315,FilaAbajo(destino));
                AgregarBit90(rotado90,FilaAbajo(destino));                
                SetPieza(FilaAbajo(destino), PeonN);
                AgregarMaterial(PeonN);
                captura = Vacio;
            }
            else if(m.Promocion()) { // Promocion
                SacarBit(PeonesB(), destino);
                AgregarMaterial(PeonB);
                switch(m.Promocion()) {
                    case DamaB:
                        SacarBit(DamasB(), destino);
                        QuitarMaterial(DamaB);
                        break;
                    case AlfilB:
                        SacarBit(AlfilesB(), destino);
                        QuitarMaterial(AlfilB);
                        break;
                    case CaballoB:
                        SacarBit(CaballosB(), destino);
                        QuitarMaterial(CaballoB);
                        break;
                    case TorreB:
                        SacarBit(TorresB(), destino);
                        QuitarMaterial(TorreB);
                        break;
                }
            }
            break;
        case PeonN:
            XOR(piezasNegras, M);
            XOR(PeonesN(), M);
            SetPieza(origen, PeonN);
            if(m.EsAP()) {
                AgregarBit(PeonesB(), FilaArriba(destino)); // Captura al paso
                AgregarBit(piezasBlancas, FilaArriba(destino));
                AgregarBit45(rotado45, FilaArriba(destino));
                AgregarBit315(rotado315,FilaArriba(destino));
                AgregarBit90(rotado90,FilaArriba(destino));                
                SetPieza(FilaArriba(destino), PeonB);
                AgregarMaterial(PeonB);
                captura = Vacio;
            }
            else if(m.Promocion()) { // Promocion
                SacarBit(PeonesN(), destino);
                AgregarMaterial(PeonN);
                switch(m.Promocion()) {
                    case DamaN:
                        SacarBit(DamasN(), destino);
                        QuitarMaterial(DamaN);
                        break;
                    case AlfilN:
                        SacarBit(AlfilesN(), destino);
                        QuitarMaterial(AlfilN);
                        break;
                    case CaballoN:
                        SacarBit(CaballosN(), destino);
                        QuitarMaterial(CaballoN);
                        break;
                    case TorreN:
                        SacarBit(TorresN(), destino);
                        QuitarMaterial(TorreN);
                        break;
                }
            }
            break;
        case CaballoB:
            XOR(piezasBlancas, M);
            XOR(CaballosB(), M);
            SetPieza(origen, CaballoB);
            break;
        case CaballoN:
            XOR(piezasNegras, M);
            XOR(CaballosN(), M);
            SetPieza(origen, CaballoN);
            break;
        case AlfilB:
            XOR(piezasBlancas, M);
            XOR(AlfilesB(), M);
            SetPieza(origen, AlfilB);
            break;
        case AlfilN:
            XOR(piezasNegras, M);
            XOR(AlfilesN(), M);
            SetPieza(origen, AlfilN);
            break;
        case TorreB:
            XOR(piezasBlancas, M);
            XOR(TorresB(), M);
            SetPieza(origen, TorreB);
            break;
        case TorreN:
            XOR(piezasNegras, M);
            XOR(TorresN(), M);
            SetPieza(origen, TorreN);
            break;
        case DamaB:
            XOR(piezasBlancas, M);
            XOR(DamasB(), M);
            SetPieza(origen, DamaB);
            break;
        case DamaN:
            XOR(piezasNegras, M);
            XOR(DamasN(), M);
            SetPieza(origen, DamaN);
            break;
    }

    // Si hubo una captura hace algunos cambios adicionales
    if(captura) {
        AgregarBit45(rotado45,destino);
        AgregarBit315(rotado315,destino);
        AgregarBit90(rotado90,destino);
        AgregarMaterial(captura);
        SetPieza(destino, captura);
    
        switch(captura) {
            case PeonB:
                AgregarBit(PeonesB(), destino);
                AgregarBit(piezasBlancas, destino);
                break;
            case PeonN:
                AgregarBit(PeonesN(), destino);
                AgregarBit(piezasNegras, destino);
                break;
            case CaballoB:
                AgregarBit(CaballosB(), destino);
                AgregarBit(piezasBlancas, destino);
                break;
            case CaballoN:
                AgregarBit(CaballosN(), destino);
                AgregarBit(piezasNegras, destino);
                break;
            case AlfilB:
                AgregarBit(AlfilesB(), destino);
                AgregarBit(piezasBlancas, destino);
                break;
            case AlfilN:
                AgregarBit(AlfilesN(), destino);
                AgregarBit(piezasNegras, destino);
                break;
            case TorreB:
                AgregarBit(TorresB(), destino);
                AgregarBit(piezasBlancas, destino);
                break;
            case TorreN:
                AgregarBit(TorresN(), destino);
                AgregarBit(piezasNegras, destino);
                break;
            case DamaB:
                AgregarBit(DamasB(), destino);
                AgregarBit(piezasBlancas, destino);
                break;
            case DamaN:
                AgregarBit(DamasN(), destino);
                AgregarBit(piezasNegras, destino);
                break;
        }
    }
#if defined(CHEQUEO_CONSISTENCIA)
    ChequearConsistencia();
#endif
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ CHEQUEO AMENAZAS
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Devuelve true si la casilla c est atacada por el lado elegido
bool Tablero::EsAtacado(Casilla c, Color ladoAtaca) const {
    if(ladoAtaca == Blancas) {
        if(AtaquesCaballo(c) & CaballosB()) return true;
        if(AtaquesRey(c) & (ReyesB()|DamasB())) return true;
        if(AtaquesPeonN(c) & PeonesB()) return true; // Parece error pero no es, pensarlo
        if(AtaquesTorre(c) & TorresB()) return true;
        if(AtaquesAlfil(c) & AlfilesB()) return true;
        if(AtaquesDama(c) & DamasB()) return true;
    }
    else {
        if(AtaquesCaballo(c) & CaballosN()) return true;
        if(AtaquesRey(c) & (ReyesN()|DamasN())) return true;
        if(AtaquesPeonB(c) & PeonesN()) return true;  // Idem anterior
        if(AtaquesTorre(c) & TorresN()) return true;
        if(AtaquesAlfil(c) & AlfilesN()) return true;
        if(AtaquesDama(c) & DamasN()) return true;
    }
    return false;
}

// Para saber si el rey est en jaque
bool Tablero::EnJaque(void) const {
    Bitboard destinos;
    if(estado.BlancasMueven()) {
        int posRey = GetPosicionReyB();
        if(AtaquesPeonB(posRey) & PeonesN()) return true;
        if(AtaquesCaballo(posRey) & CaballosN()) return true;
        destinos = AlfilesN() | DamasN();
        if((MaskAlfil[posRey] & destinos) && (AtaquesAlfil(posRey) & destinos)) return true;
        destinos = TorresN() | DamasN();
        if((MaskTorre[posRey] & destinos) && (AtaquesTorre(posRey) & destinos)) return true;
        if(AtaquesRey(posRey) & ReyesN()) return true;
        return false;
    }
    else {
        int posRey = GetPosicionReyN();
        if(AtaquesPeonN(posRey) & PeonesB()) return true;
        if(AtaquesCaballo(posRey) & CaballosB()) return true;
        destinos = AlfilesB() | DamasB();
        if((MaskAlfil[posRey] & destinos) && (AtaquesAlfil(posRey) & destinos)) return true;
        destinos = TorresB() | DamasB();
        if((MaskTorre[posRey] & destinos) && (AtaquesTorre(posRey) & destinos)) return true;
        if(AtaquesRey(posRey) & ReyesB()) return true;
        return false;
    }
}

// Devuelve un bitboard con las posiciones que atacan a la pieza elegida
Bitboard Tablero::AtaquesCasilla(Casilla c) const {
    return ( AtaquesPeonB(c) & PeonesN() | AtaquesPeonN(c) & PeonesB() |
             AtaquesCaballo(c) & (CaballosB() | CaballosN()) |
             AtaquesAlfil(c) & AlfilesDamas() |
             AtaquesTorre(c) & TorresDamas() |
             AtaquesRey(c) & (ReyesB() | ReyesN()));
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ SEE
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int Tablero::SEE(Movimiento m) const {
    Bitboard ataques;
    int piezaAtacada;
    int listaCapturas[32];
    Direccion direccion;
    int numeroCapturas = 1;
    Casilla casilla;
    Color color;

    // Accesos a la informacin
    Casilla origen = m.Origen();
    Casilla destino = m.Destino();
    Color colorAtaque = ColorPieza(m.GetPieza());

    // Obtenemos las piezas que atacan la casilla
    ataques = AtaquesCasilla(destino);

    // Establecemos el valor de la pieza capturada
    piezaAtacada = VALOR_PIEZAS_SEE[GetPieza(destino)];

    // Agregamos la pieza que captura en el movimiento como la primer captura
    color = ColorContrario(colorAtaque);
    listaCapturas[0] = piezaAtacada;
    // Si el movimiento es una promocin, entonces se toma la pieza a la que promociona
    piezaAtacada = (m.Promocion()) ? VALOR_PIEZAS_SEE[m.Promocion()] : VALOR_PIEZAS_SEE[GetPieza(origen)];
    SacarBit(ataques, origen);
    direccion = Direcciones[destino][origen]; // Direccion hacia la que se descubren los ataques
    if(direccion != SIN_DIRECCION)
        ataques = EncontrarAtaquesOcultos(ataques, origen, direccion);

    // Ahora para cada lado tomamos la pieza que ataca a destino, pero la que tiene el menor valor
    // Si se encuentra una, se sigue el proceso antes viendo si hay ataques ocultos
    while(ataques) {
        if(color == Blancas) {
            // El orden ac es para elegir antes la pieza de menor valor
            if(PeonesB() & ataques)
                casilla = MSB(PeonesB() & ataques);
            else if(CaballosB() & ataques)
                casilla = MSB(CaballosB() & ataques);
            else if(AlfilesB() & ataques)
                casilla = MSB(AlfilesB() & ataques);
            else if(TorresB() & ataques)
                casilla = MSB(TorresB() & ataques);
            else if(DamasB() & ataques)
                casilla = MSB(DamasB() & ataques);
            else if(ReyesB() & ataques)
                casilla = GetPosicionReyB();
            else
                break; // No hay mas ataques blancos, termina el proceso
        }
        else {
            // El orden ac es para elegir antes la pieza de menor valor
            if(PeonesN() & ataques)
                casilla = MSB(PeonesN() & ataques);
            else if(CaballosN() & ataques)
                casilla = MSB(CaballosN() & ataques);
            else if(AlfilesN() & ataques)
                casilla = MSB(AlfilesN() & ataques);
            else if(TorresN() & ataques)
                casilla = MSB(TorresN() & ataques);
            else if(DamasN() & ataques)
                casilla = MSB(DamasN() & ataques);
            else if(ReyesN() & ataques)
                casilla = GetPosicionReyN();
            else
                break; // No hay mas ataques negros, termina el proceso
        }

        // Ahora agregamos la pieza que ataca como una captura mas, y vemos si genera ataques ocultos
        listaCapturas[numeroCapturas] = -listaCapturas[numeroCapturas-1] + piezaAtacada;
        piezaAtacada = VALOR_PIEZAS_SEE[GetPieza(casilla)];
        SacarBit(ataques, casilla);
        direccion = Direcciones[destino][casilla];
        if(direccion != SIN_DIRECCION)
            ataques = EncontrarAtaquesOcultos(ataques, casilla, direccion);
        numeroCapturas++;
        color = ColorContrario(color);
    }

    // Se extrae el valor de la secuencia de capturas, aca se ve cuando 
    // conviene detener los intercambios de piezas, asi sabemos el valor real
    while(--numeroCapturas) {
        if(listaCapturas[numeroCapturas] > -listaCapturas[numeroCapturas-1])
            listaCapturas[numeroCapturas-1] = -listaCapturas[numeroCapturas];
    }

    return listaCapturas[0];
}

// Encuentra ataques ocultos al sacar una pieza del listado de atacantes
Bitboard Tablero::EncontrarAtaquesOcultos(Bitboard &ataques, Casilla origen, Direccion direccion) const {
    switch(direccion) {
        case NORTE:
            return (ataques | (AtaquesTorreCol(origen) & TorresDamas() & MaskNorte[origen]));
        case SUR:
            return (ataques | (AtaquesTorreCol(origen) & TorresDamas() & MaskSur[origen]));
        case ESTE:
            return (ataques | (AtaquesTorreFila(origen) & TorresDamas() & MaskEste[origen]));
        case OESTE:
            return (ataques | (AtaquesTorreFila(origen) & TorresDamas() & MaskOeste[origen]));
        case NORESTE:
            return (ataques | (AtaquesAlfilDiag315(origen) & AlfilesDamas() & MaskNoreste[origen]));
        case NOROESTE:
            return (ataques | (AtaquesAlfilDiag45(origen) & AlfilesDamas() & MaskNoroeste[origen]));
        case SUDESTE:
            return (ataques | (AtaquesAlfilDiag45(origen) & AlfilesDamas() & MaskSudeste[origen]));
        case SUDOESTE:
            return (ataques | (AtaquesAlfilDiag315(origen) & AlfilesDamas() & MaskSudoeste[origen]));
        default:
            return ataques;
    }
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ CLAVADAS
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Chequea si una pieza blanca en la casilla est clavada por defender el rey blanco
bool Tablero::ClavadaReyBlanco(Casilla casilla) const {
    Casilla posRey = GetPosicionReyB();

    // El rey y la casilla no estn en una hilera
    if (Direcciones[posRey][casilla] == SIN_DIRECCION || MaskEntre[posRey][casilla] & Ocupadas())
        return 0;

    switch (Direcciones[posRey][casilla]) {
    case NORTE: case SUR:
		// Ataques sobre hileras verticales
		if(AtaquesTorreCol(casilla) & ReyesB())
			return (bool)((AtaquesTorreCol(casilla) & (TorresN() | DamasN())) != 0);
		else
			return false;
        break;

	case ESTE: case OESTE:
		// Ataques sobre hileras horizontales
		if(AtaquesTorreFila(casilla) & ReyesB())
			return (bool)((AtaquesTorreFila(casilla) & (TorresN() | DamasN())) != 0);
		else
			return false;
        break;

    case NOROESTE: case SUDESTE:
		// Ataques sobre las diagonales A8-H1
		if(AtaquesAlfilDiag45(casilla) & ReyesB())
			return (bool)((AtaquesAlfilDiag45(casilla) & (AlfilesN() | DamasN())) != 0);
		else
			return false;
        break;

    case NORESTE: case SUDOESTE:
		// Ataques sobre las diagonales A1-H8
		if(AtaquesAlfilDiag315(casilla) & ReyesB())
			return (bool)((AtaquesAlfilDiag315(casilla) & (AlfilesN() | DamasN())) != 0);
		else
			return false;
        break;
    }

    return false;
}

// Chequea si una pieza negra en la casilla est clavada por defender el rey negro
bool Tablero::ClavadaReyNegro(Casilla casilla) const {
    Casilla posRey = GetPosicionReyN();

    // El rey y la casilla no estn en una hilera
    if (Direcciones[posRey][casilla] == SIN_DIRECCION || MaskEntre[posRey][casilla] & Ocupadas())
        return 0;

    switch (Direcciones[posRey][casilla]) {
    case NORTE: case SUR:
		// Ataques sobre hileras verticales
		if(AtaquesTorreCol(casilla) & ReyesN())
			return (bool)((AtaquesTorreCol(casilla) & (TorresB() | DamasB())) != 0);
		else
			return false;
        break;

	case ESTE: case OESTE:
		// Ataques sobre hileras horizontales
		if(AtaquesTorreFila(casilla) & ReyesN())
			return (bool)((AtaquesTorreFila(casilla) & (TorresB() | DamasB())) != 0);
		else
			return false;
        break;

    case NOROESTE: case SUDESTE:
		// Ataques sobre las diagonales A8-H1
		if(AtaquesAlfilDiag45(casilla) & ReyesN())
			return (bool)((AtaquesAlfilDiag45(casilla) & (AlfilesB() | DamasB())) != 0);
		else
			return false;
        break;

    case NORESTE: case SUDOESTE:
		// Ataques sobre las diagonales A1-H8
		if(AtaquesAlfilDiag315(casilla) & ReyesN())
			return (bool)((AtaquesAlfilDiag315(casilla) & (AlfilesB() | DamasB())) != 0);
		else
			return false;
        break;
    }

    return false;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ DEBUG MOSTRAR
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Devuelve una lista con los movimientos legales en un string
string Tablero::ListaMovimientosLegales(void) const {
    Tablero tableroCopia(*this);
    ListaMovimientos listaMovimientos(&tableroCopia);
    Movimiento m;
    string ret = "";

    listaMovimientos.SetTipoLista(PERFT);
    while((m = listaMovimientos.GetSiguiente()).EsNulo() == false) {
        tableroCopia.Hacer(m);
        if(tableroCopia.PosicionLegal())
            ret += Notador::ObtenerNotacion(m) + " ";
        tableroCopia.Deshacer();
    }

    return ret;
}

// Devuelve la cantidad de movimientos legales disponibles
int Tablero::TotalMovimientosLegales(void) const {
    Tablero tableroCopia(*this);
    ListaMovimientos listaMovimientos(&tableroCopia);
    Movimiento m;

    int cantidad = 0;
    listaMovimientos.SetTipoLista(PERFT);
    while((m = listaMovimientos.GetSiguiente()).EsNulo() == false) {
        tableroCopia.Hacer(m);
        if(tableroCopia.PosicionLegal())
            cantidad++;
        tableroCopia.Deshacer();
    }

    return cantidad;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ HASH KEY
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Genera la hashkey para la posicion actual
Hashkey Tablero::GenerarHashkey(void) const {
    Hashkey hash = 0;

    for(Casilla c = A1; c <= H8; c++) {
        if(GetPieza(c) != Vacio)
            ActualizarHash(hash, GetPieza(c), c);
    }
    
    ActualizarHashAlPaso(hash, estado.GetCasillaAlPaso());
    ActualizarHashEnroque(hash, estado.GetInfoEnroque());
    if(!estado.BlancasMueven())
        ActualizarHashLado(hash);

    return hash;    
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ CONTEO REPETICIONES
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Devuelve el conteo de repeticiones en la busqueda
int Tablero::ConteoRepeticionesEnBusqueda(int ply) const {
    int reps = (ply > 0) ? 1 : 0; // Si ply > 0 la posicion actual es parte de la busqueda
    int hply = historial.GetPly();
    
    for(int i = hply-2; i >= 0 && i >= hply-estado.GetCincuenta() && i > (hply-ply+1); i-=2) {
        if(historial.GetHashKey(i) == hashkey)
            reps++;
    }
    
    return reps;
}

// Devuelve el conteo de repeticiones en la historia
int Tablero::ConteoRepeticionesEnHistoria(int ply) const {
    int reps = 0;
    int hply = historial.GetPly();
    
    for(int i = hply-ply+1; i >= 0 && i >= hply-estado.GetCincuenta(); i-=2) {
        if(historial.GetHashKey(i) == hashkey)
            reps++;
    }
    
    return reps;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ VALIDEZ MOVIMIENTOS
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Chequea si un movimiento es pseudo-legal para la posicion
bool Tablero::EsMovimientoValido(Movimiento mov) const {
    Casilla origen = mov.Origen();
    Casilla destino = mov.Destino();
    Pieza pieza = mov.GetPieza();

    // Posibles salidas tempranas
    if(mov.EsNulo()) return false;
    if(GetPieza(origen) != pieza) return false;
    if(ColorPieza(pieza) != estado.GetLadoMueve()) return false;
    if(mov.Captura() != Vacio && ColorPieza(mov.Captura()) != ColorContrario(ColorPieza(pieza))) return false;
    
    // Segn la pieza que mueve vemos que hacemos    
    switch(pieza) {
        // Peones blancos
        case PeonB:
            if(destino == FilaArriba(origen)) {
                if(GetPieza(destino) == Vacio)
                    return true;
                return false;
            }
            else if(destino == DosFilasArriba(origen)) {
                if(GetPieza(destino) == Vacio && GetPieza(FilaArriba(origen)) == Vacio)
                    return true;
                return false;
            }
            // Capturas
            if(!mov.Captura()) // Si es captura normal no necesitamos chequear
                return false;            
            // Captura al paso
            if(mov.EsAP()) {
                if(GetPieza(FilaAbajo(destino)) == PeonN && estado.GetCasillaAlPaso() == destino)
                    return true;
                return false;
            }
            break;
        // Peones negros
        case PeonN:
            if(destino == FilaAbajo(origen)) {
                if(GetPieza(destino) == Vacio)
                    return true;
                return false;
            }
            else if(destino == DosFilasAbajo(origen)) {
                if(GetPieza(destino) == Vacio && GetPieza(FilaAbajo(origen)) == Vacio)
                    return true;
                return false;
            }
            // Capturas
            if(!mov.Captura()) // Si es captura normal no necesitamos chequear
                return false;            
            // Captura al paso
            if(mov.EsAP()) {
                if(GetPieza(FilaArriba(destino)) == PeonB && estado.GetCasillaAlPaso() == destino)
                    return true;
                return false;
            }
            break;
        // Reyes blancos
        case ReyB:
            if(mov.EsECB()) {
                if(!estado.GetInfoEnroque().BlancasPuedenEnrocarCorto()) return false;
                if(Ocupadas()&(Mask[F1]|Mask[G1]) || EsAtacado(E1,Negras) || EsAtacado(F1,Negras))
                    return false;
                return true;
            }
            else if(mov.EsELB()) {
                if(!estado.GetInfoEnroque().BlancasPuedenEnrocarLargo()) return false;
                if(Ocupadas()&(Mask[C1]|Mask[D1]) || EsAtacado(E1,Negras) || EsAtacado(D1,Negras))
                    return false;
                return true;                
            }
            break;
        // Reyes negros
        case ReyN:
            if(mov.EsECN()) {
                if(!estado.GetInfoEnroque().NegrasPuedenEnrocarCorto()) return false;
                if(Ocupadas()&(Mask[F8]|Mask[G8]) || EsAtacado(E8,Blancas) || EsAtacado(F8,Blancas))
                    return false;
                return true;
            }
            else if(mov.EsELN()) {
                if(!estado.GetInfoEnroque().BlancasPuedenEnrocarLargo()) return false;
                if(Ocupadas()&(Mask[C8]|Mask[D8]) || EsAtacado(E8,Blancas) || EsAtacado(D8,Blancas))
                    return false;
                return true;                
            }
            break;
        // Piezas deslizantes
        case DamaB:
        case DamaN:
        case TorreB:
        case TorreN:
        case AlfilB:
        case AlfilN:
            if(!(MaskEntre[origen][destino] & Ocupadas()))
                break;
            return false;
        // Caballos, por la forma que se mueven, solo comprobamos la captura
        case CaballoB:
        case CaballoN:
            break;
    }

    // Si saltamos hasta ac vamos bien, veamos si la captura corresponde y no es un rey
    if(mov.Captura() == GetPieza(destino) && mov.Captura() != ReyB && mov.Captura() != ReyN)
        return true;

    return false;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ INSUFICIENCIA DE MATERIAL
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

bool Tablero::InsuficienciaMaterialMate(void) const {
    int total = PopCount(Ocupadas());

    if(total == 2) {
        // Los reyes solos es tablas
        return true;
    }
    else if(total == 3) {
        // Rey y una pieza menor contra rey es tablas
        if(PopCount(CaballosB() | CaballosN() | AlfilesB() | AlfilesN()) == 1)
            return true;
    }
    else if(total == 4) {
        // Rey y una pieza menor contra rey y una pieza menor
        if(PopCount(CaballosB() | AlfilesB()) == 1 && PopCount(CaballosN() | AlfilesN()) == 1)
            return true;
        // Rey y dos caballos contra rey "se puede" considerar tablas pues no se pude forzar el mate
        // se requiere para ello que el oponente cometa un gran error
        if(PopCount(CaballosB() == 2) || PopCount(CaballosN()) == 2)
            return true;
    }
    else if(total == 5) {
        // Rey y dos caballos contra rey y una pieza menor "se puede" considerar tablas (ver anterior)
        if( (PopCount(CaballosB() == 2) && PopCount(CaballosN() | AlfilesN()) == 1) || (PopCount(CaballosN() == 2) && PopCount(CaballosB() | AlfilesB()) == 1) )
            return true;
    }
    
    return false;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ CHEQUEO DE CONSISTENCIA
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int Tablero::ChequearConsistencia(void) const {
    int codigoError = 0;
#if defined(CHEQUEO_CONSISTENCIA)
	Bitboard peonesB,peonesN,caballosB,caballosN,alfilesB,alfilesN,torresB;
    Bitboard torresN,damasB,damasN,blancas,negras,r90,r45,r315;
    Bitboard posReyB,posReyN;
    Hashkey hash;
    int materialB, materialN;
    int valorPiezasB, valorPiezasN;
    int totalPeonesB, totalPeonesN;
    Pieza pieza;

    // Ponemos todo a cero
	peonesB = caballosB = alfilesB = torresB = damasB = 0;
	peonesN = caballosN = alfilesN = torresN = damasN = 0;
	blancas = negras = 0;
	r90 = r45 = r315 = 0;
	posReyB = posReyN = 0;
    materialB = materialN = 0;
    valorPiezasB = valorPiezasN = 0;
    totalPeonesB = totalPeonesN = 0;
    hash = 0;

    // Recorremos el tablero obteniendo la informacin  
    for(Casilla i = A1; i <= H8; i++) {
        pieza = GetPieza(i);
        if(pieza != Vacio) {
            AgregarBit90(r90,i);
            AgregarBit45(r45,i);
            AgregarBit315(r315,i);
            ActualizarHash(hash, pieza, i);
            if(ColorPieza(pieza) == Blancas) {
                AgregarBit(blancas,i);
                materialB += VALOR_MATERIAL[pieza];
                valorPiezasB+=VALOR_PIEZAS[pieza]; // No toma en cuenta peones ni reyes
                switch(pieza) {
                    case PeonB:
                        AgregarBit(peonesB,i);
                        totalPeonesB++;
                        break;
                    case CaballoB:
                        AgregarBit(caballosB,i);
                        break;
                    case AlfilB:
                        AgregarBit(alfilesB,i);
                        break;
                    case TorreB:
                        AgregarBit(torresB,i);
                        break;
                    case DamaB:
                        AgregarBit(damasB,i);
                        break;
                    case ReyB:
                        posReyB = i;
                        break;
                }
            }
            else {
                AgregarBit(negras,i);
                materialN += VALOR_MATERIAL[pieza];
                valorPiezasN+=VALOR_PIEZAS[pieza]; // No toma en cuenta peones ni reyes
                switch(pieza) {
                    case PeonN:
                        AgregarBit(peonesN,i);
                        totalPeonesN++;
                        break;
                    case CaballoN:
                        AgregarBit(caballosN,i);
                        break;
                    case AlfilN:
                        AgregarBit(alfilesN,i);
                        break;
                    case TorreN:
                        AgregarBit(torresN,i);
                        break;
                    case DamaN:
                        AgregarBit(damasN,i);
                        break;
                    case ReyN:
                        posReyN = i;
                        break;
                }
            }
        }
    }
    
    ActualizarHashAlPaso(hash, GetEstado().GetCasillaAlPaso());
    ActualizarHashEnroque(hash, GetEstado().GetInfoEnroque());
    if(!GetEstado().BlancasMueven())
        ActualizarHashLado(hash);

    // Chequeo de errores posicion reyes
    if(posReyB != GetPosicionReyB())
		codigoError |= 0x01;
    if(posReyN != GetPosicionReyN())
		codigoError |= 0x02;

    // Chequeo errores bitboards piezas
    if(blancas != piezasBlancas)
        codigoError |= 0x04;
    if(negras != piezasNegras)
        codigoError |= 0x08;
	if(peonesB != PeonesB())
		codigoError |= 0x10;
	if(caballosB != CaballosB())
		codigoError |= 0x20;
	if(alfilesB != AlfilesB())
		codigoError |= 0x40;
	if(torresB != TorresB())
		codigoError |= 0x80;
	if(damasB != DamasB())
		codigoError |= 0x100;
	if(peonesN != PeonesN())
		codigoError |= 0x200;
	if(caballosN != CaballosN())
		codigoError |= 0x400;
	if(alfilesN != AlfilesN())
		codigoError |= 0x800;
	if(torresN != TorresN())
		codigoError |= 0x1000;
	if(damasN != DamasN())
		codigoError |= 0x2000;
    
    // Chequeo errores bitboards rotados
    if(r90 != rotado90)
        codigoError |= 0x4000;
    if(r45 != rotado45)
        codigoError |= 0x8000;
    if(r315 != rotado315)
        codigoError |= 0x10000;

    // Chequeo errores material
    if(materialB != material[Blancas])
        codigoError |= 0x20000;
    if(materialN != material[Negras])
        codigoError |= 0x40000;

    // Chequeo errores valor piezas
    if(valorPiezasB != totalValorPiezas[Blancas])
        codigoError |= 0x80000;
    if(valorPiezasN != totalValorPiezas[Negras])
        codigoError |= 0x100000;

    // Chequeo errores total peones
    if(totalPeonesB != totalPeones[Blancas])
        codigoError |= 0x200000;
    if(totalPeonesN != totalPeones[Negras])
        codigoError |= 0x400000;

    // Chequeo hashkey
    if(hash != hashkey)
        codigoError |= 0x800000;

    // Mostramos los resultados
    if(codigoError)
        tableroDebug->MostrarCodigoError(codigoError);
#endif
    return codigoError;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ FUNCIONES INLINE
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Establece la posicin del rey blanco
inline void Tablero::SetPosicionReyB(Casilla c) {
    assert(CasillaValida(c));
    posReyB = c;
}

// Establece la posicin del rey negro
inline void Tablero::SetPosicionReyN(Casilla c) {
    assert(CasillaValida(c));
    posReyN = c;
}

// Establece una pieza en una casilla seleccionada
inline void Tablero::SetPieza(Casilla casilla, Pieza pieza) {
    assert(CasillaValida(casilla));
    assert(PiezaValida(pieza));
    casillas[casilla] = pieza;
}

// Devuelve la pieza en la casilla seleccionada
inline Pieza Tablero::GetPieza(Casilla casilla) const {
    assert(CasillaValida(casilla));
    return casillas[casilla];
}

// Agrega la pieza al material
inline void Tablero::AgregarMaterial(Pieza pieza) {
    assert(PiezaValida(pieza));
    material[ColorPieza(pieza)] += VALOR_MATERIAL[pieza];
    AgregarPieza(pieza);
    AgregarPeon(pieza);
}

// Quita la pieza de material
inline void Tablero::QuitarMaterial(Pieza pieza) {
    assert(PiezaValida(pieza));
    material[ColorPieza(pieza)] -= VALOR_MATERIAL[pieza];
    QuitarPieza(pieza);
    QuitarPeon(pieza);
}

// Agrega una pieza al conteo del valor
inline void Tablero::AgregarPieza(Pieza pieza) {
    assert(PiezaValida(pieza));
    totalValorPiezas[ColorPieza(pieza)] += VALOR_PIEZAS[pieza];
}

// Quita una pieza al conteo del valor
inline void Tablero::QuitarPieza(Pieza pieza) {
    assert(PiezaValida(pieza));
    totalValorPiezas[ColorPieza(pieza)] -= VALOR_PIEZAS[pieza];
}

// Agrega un peon al conteo de peones
inline void Tablero::AgregarPeon(Pieza peon) {
    assert(PiezaValida(peon));
    totalPeones[ColorPieza(peon)] += VALOR_PEONES[peon];
}

// Quita un peon al conteo de peones
inline void Tablero::QuitarPeon(Pieza peon) {
    assert(PiezaValida(peon));
    totalPeones[ColorPieza(peon)] -= VALOR_PEONES[peon];
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ ATAQUES PARA GENERACION MOVIMIENTOS
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

inline Bitboard Tablero::EstadoFila(int c) const {
    return (((piezasBlancas|piezasNegras)>>((c&~7)+1))&63);
}

inline Bitboard Tablero::EstadoColumna(int c) const {
    return ((rotado90>>(((c&7)<<3)+1))&63);
}

inline Bitboard Tablero::EstadoDiag45(int c) const {
    return ((rotado45>>Bitboards::Shift45[c])&63);
}

inline Bitboard Tablero::EstadoDiag315(int c) const {
    return ((rotado315>>Bitboards::Shift315[c])&63);
}

inline Bitboard Tablero::AtaquesTorreFila(int c) const {
    return Bitboards::AtaquesFila[c][EstadoFila(c)];
}

inline Bitboard Tablero::AtaquesTorreCol(int c) const {
    return Bitboards::AtaquesColumna[c][EstadoColumna(c)];
}

inline Bitboard Tablero::AtaquesAlfilDiag45(int c) const {
    return Bitboards::AtaquesDiag45[c][EstadoDiag45(c)];
}

inline Bitboard Tablero::AtaquesAlfilDiag315(int c) const {
    return Bitboards::AtaquesDiag315[c][EstadoDiag315(c)];
}

inline Bitboard Tablero::AtaquesTorre(int c) const {
    return (AtaquesTorreFila(c)|AtaquesTorreCol(c));
}

inline Bitboard Tablero::AtaquesAlfil(int c) const {
    return (AtaquesAlfilDiag45(c)|AtaquesAlfilDiag315(c));
}

inline Bitboard Tablero::AtaquesDama(int c) const {
    return (AtaquesTorre(c)|AtaquesAlfil(c));
}

inline Bitboard Tablero::AtaquesCaballo(int c) const {
    return (Bitboards::AtaquesCaballos[c]);
}

inline Bitboard Tablero::AtaquesRey(int c) const {
    return (Bitboards::AtaquesReyes[c]);
}

inline Bitboard Tablero::AtaquesPeonB(int c) const {
    return (Bitboards::AtaquesPeonesB[c]);
}

inline Bitboard Tablero::AtaquesPeonN(int c) const {
    return (Bitboards::AtaquesPeonesN[c]);
}
