//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ Tablero.h
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#pragma once

#include "Bitboards.h"
#include "Hashkeys.h"
#include "Historial.h"

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ DECLARACION PREVIA DE CLASES
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class FEN;
class ListaMovimientos;
class TableroDebug;

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ class Tablero
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class Tablero {
public:
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ AMISTAD
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    
    friend class TableroDebug;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ CONSTRUCTORES
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Constructor
    Tablero(void);
    // Constructor copia
    Tablero(const Tablero &tablero);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ EL ANALIZADOR DEL TABLERO
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    void SetTableroDebug(const TableroDebug *td);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ FUNCIONES INLINE
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    bool PosicionLegal(void);

    EstadoTablero GetEstado(void) const;
    EstadoTablero& GetEstado(void);
    Hashkey GetHashkey(void) const;
    int DiferenciaMaterial(void) const;
    Casilla GetPosicionReyB(void) const;
    Casilla GetPosicionReyN(void) const;
    int GetTotalValorPiezas(Color color) const;
    int GetTotalPeones(Color color) const;

    void HacerMovimientoNulo(void);
    void DeshacerMovimientoNulo(void);

    const Historial& GetHistorial(void) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ HACER Y DESHACER
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Realiza un movimiento en el tablero
    void Hacer(Movimiento m);

    // Deshace el ultimo movimiento realizado sobre el tablero
    void Deshacer(void);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ ESTABLECIMIENTO DE POSICIONES
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Configura el tablero segn una cadena FEN
    bool LeerFEN(FEN &fen);

    // Establece la posicion inicial de una partida de ajedrez
    void EstablecerPosicionInicial(void);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ GENERACION DE MOVIMIENTOS
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Genera capturas y promociones
    void GenerarCapturas(ListaMovimientos &listaMovs) const;

    // Genera no capturas: movimientos simples y enroques
    void GenerarNoCapturas(ListaMovimientos &listaMovs) const;

    // Genera evasiones legales al jaque
    void GenerarEvasiones(ListaMovimientos &listaMovs) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ ATAQUES Y JAQUES
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Devuelve true si la casilla c est atacada por el lado elegido
    bool EsAtacado(Casilla c, Color ladoAtaca) const;

    // Para saber si el rey est en jaque
    bool EnJaque(void) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ LEGALIDAD MOVIMIENTOS
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Devuelve una lista de los movimientos legales disponibles en un string   
    string ListaMovimientosLegales(void) const;

    // Devuelve la cantidad de movimientos legales disponibles
    int Tablero::TotalMovimientosLegales(void) const;

    // Devuelve true si el movimiento sirve para la posicion (pseudo-legalidad)
    bool EsMovimientoValido(Movimiento mov) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ CONTEO REPETICIONES
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Devuelve el conteo de repeticiones en la busqueda
    int ConteoRepeticionesEnBusqueda(int ply) const;

    // Devuelve el conteo de repeticiones en la historia
    int ConteoRepeticionesEnHistoria(int ply) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ STATIC EXCHANGE EVALUATOR
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // SEE: Static Exchange Evaluator
    int SEE(Movimiento m) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ MATERIAL PARA MATE
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Devuelve true si no hay material en el tablero para que algun lado haga jaque mate
    bool InsuficienciaMaterialMate(void) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ FUNCIONES ACCESO BITBOARDS
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Versiones constantes
    Bitboard Ocupadas(void) const;
    Bitboard Vacias(void) const;
    Bitboard PeonesB(void) const;
    Bitboard PeonesN(void) const;
    Bitboard CaballosB(void) const;
    Bitboard CaballosN(void) const;
    Bitboard AlfilesB(void) const;
    Bitboard AlfilesN(void) const;
    Bitboard TorresB(void) const;
    Bitboard TorresN(void) const;
    Bitboard DamasB(void) const;
    Bitboard DamasN(void) const;
    Bitboard ReyesB(void) const;
    Bitboard ReyesN(void) const;
    Bitboard TorresDamas(void) const;
    Bitboard AlfilesDamas(void) const;

private:

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ FUNCIONES INLINE
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    void SetPosicionReyB(Casilla c);
    void SetPosicionReyN(Casilla c);
    void SetPieza(Casilla casilla, Pieza pieza);
    Pieza GetPieza(int fila, int col) const;
    Pieza GetPieza(Casilla casilla) const;
    void AgregarMaterial(Pieza pieza);
    void QuitarMaterial(Pieza pieza);
    void AgregarPieza(Pieza pieza);
    void QuitarPieza(Pieza pieza);
    void AgregarPeon(Pieza peon);
    void QuitarPeon(Pieza peon);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ FUNCIONES NO INLINE
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Regenera los bitboards, material y otras estructuras que se actualizan automaticamente
    void RegenerarEstructuras(void);

    // Limpia el tablero, dejando todo en blanco
    void Limpiar(void);

    // Devuelve un bitboard con las posiciones que atacan a la pieza elegida
    Bitboard AtaquesCasilla(Casilla c) const;

    // Genera la hashkey para la posicion actual
    Hashkey GenerarHashkey(void) const;

    // Encuentra ataques ocultos para la SEE
    Bitboard EncontrarAtaquesOcultos(Bitboard &ataques, Casilla origen, Direccion direccion) const;

    // Chequea si una pieza blanca en la casilla est clavada por defender el rey blanco
    bool ClavadaReyBlanco(Casilla casilla) const;

    // Chequea si una pieza negra en la casilla est clavada por defender el rey negro
    bool ClavadaReyNegro(Casilla casilla) const;

    // Chequea la consistencia de la posicion
    int ChequearConsistencia(void) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ FUNCIONES ACCESO BITBOARDS
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // Versiones que devuelven referencia para ser modificados
    Bitboard& PeonesB(void);
    Bitboard& PeonesN(void);
    Bitboard& CaballosB(void);
    Bitboard& CaballosN(void);
    Bitboard& AlfilesB(void);
    Bitboard& AlfilesN(void);
    Bitboard& TorresB(void);
    Bitboard& TorresN(void);
    Bitboard& DamasB(void);
    Bitboard& DamasN(void);
    Bitboard& ReyesB(void);
    Bitboard& ReyesN(void);

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ ATAQUES PARA GENERACION MOVIMIENTOS
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    Bitboard EstadoFila(int c) const;
    Bitboard EstadoColumna(int c) const;
    Bitboard EstadoDiag45(int c) const;
    Bitboard EstadoDiag315(int c) const;
    
    Bitboard AtaquesTorreFila(int c) const;
    Bitboard AtaquesTorreCol(int c) const;
    Bitboard AtaquesAlfilDiag45(int c) const;
    Bitboard AtaquesAlfilDiag315(int c) const;
    
    Bitboard AtaquesTorre(int c) const;
    Bitboard AtaquesAlfil(int c) const;
    Bitboard AtaquesDama(int c) const;
    Bitboard AtaquesCaballo(int c) const;
    Bitboard AtaquesRey(int c) const;
    Bitboard AtaquesPeonB(int c) const;
    Bitboard AtaquesPeonN(int c) const;

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //+ ATRIBUTOS
    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    // El Tablero
    Pieza           casillas[64];

    // Estado del tablero
    EstadoTablero   estado;
    Hashkey         hashkey;

    // Para ahorrar procesos
    int             material[2];
    int             totalValorPiezas[2];
    int             totalPeones[2];
    Casilla         posReyB;
    Casilla         posReyN; 

    // El historial de movimientos
    Historial       historial;

    // Los bitboards para la posicin
    Bitboard        piezasBlancas;
    Bitboard        piezasNegras;
    Bitboard        piezas[MAX_PIEZAS];
    Bitboard        rotado90;
    Bitboard        rotado45;
    Bitboard        rotado315;

    // El analizador del tablero
    const TableroDebug    *tableroDebug;
};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ FUNCIONES INLINE
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Chequea la legalidad de la posicion al ver si el rey contrario est en jaque
inline bool Tablero::PosicionLegal(void) {
    estado.CambiarLado();
    bool ret = !EnJaque();
    estado.CambiarLado();
    return ret;
}

// Devuelve el estado del tablero (const)
inline EstadoTablero Tablero::GetEstado(void) const {
    return estado;
}

// Devuelve el estado del tablero (no const)
inline EstadoTablero& Tablero::GetEstado(void){
    return estado;
}

// Devuelve la pieza en las coordenadas seleccionadas
inline Pieza Tablero::GetPieza(int fila, int col) const {
    assert(FilaValida(fila));
    assert(ColumnaValida(col));
    return casillas[Coord(fila, col)];
}

// Devuelve la hashkey de la posicin
inline Hashkey Tablero::GetHashkey(void) const {
    return hashkey;
}

// Devuelve la diferencia de material
inline int Tablero::DiferenciaMaterial(void) const {
    return (material[Blancas] - material[Negras]);
}

// Devuelve la posicin del rey blanco
inline Casilla Tablero::GetPosicionReyB(void) const {
    return posReyB;
}

// Devuelve la posicin del rey negro
inline Casilla Tablero::GetPosicionReyN(void) const {
    return posReyN;
}

// Devuelve el total del valor de las piezas de un determinado color
inline int Tablero::GetTotalValorPiezas(Color color) const {
    assert(ColorValido(color));
    return totalValorPiezas[color];
}

// Devuelve el total de peones de un determinado color
inline int Tablero::GetTotalPeones(Color color) const {
    assert(ColorValido(color));
    return totalPeones[color];
}

// Hace un movimiento nulo en el tablero
inline void Tablero::HacerMovimientoNulo(void) {
    historial.AgregarNulo(GetEstado());
    historial.AumentarPly();
    estado.IncrementarCincuenta();
    estado.CambiarLado();
    estado.SetCasillaAlPaso(Invalida);
    Hashkeys::ActualizarHashMovimientoNulo(hashkey);
}

// Deshace un movimiento nulo hecho en el tablero
inline void Tablero::DeshacerMovimientoNulo(void) {
    Hashkeys::ActualizarHashMovimientoNulo(hashkey);
    historial.ReducirPly();
    estado = historial.GetEstadoTablero();
}

// Devuelve el historial del tablero
inline const Historial& Tablero::GetHistorial(void) const {
    return historial;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+ FUNCIONES ACCESO BITBOARDS
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Versiones constantes-publicas
inline Bitboard Tablero::Ocupadas(void) const { return (piezasBlancas|piezasNegras); }
inline Bitboard Tablero::Vacias(void) const { return ~(piezasBlancas|piezasNegras); }
inline Bitboard Tablero::PeonesB(void) const { return piezas[PeonB]; }
inline Bitboard Tablero::PeonesN(void) const { return piezas[PeonN]; }
inline Bitboard Tablero::CaballosB(void) const { return piezas[CaballoB]; }
inline Bitboard Tablero::CaballosN(void) const { return piezas[CaballoN]; }
inline Bitboard Tablero::AlfilesB(void) const { return piezas[AlfilB]; }
inline Bitboard Tablero::AlfilesN(void) const { return piezas[AlfilN]; }
inline Bitboard Tablero::TorresB(void) const { return piezas[TorreB]; }
inline Bitboard Tablero::TorresN(void) const { return piezas[TorreN]; }
inline Bitboard Tablero::DamasB(void) const { return piezas[DamaB]; }
inline Bitboard Tablero::DamasN(void) const { return piezas[DamaN]; }
inline Bitboard Tablero::ReyesB(void) const { return Bitboards::Mask[posReyB]; }
inline Bitboard Tablero::ReyesN(void) const { return Bitboards::Mask[posReyN]; }
inline Bitboard Tablero::TorresDamas(void) const { return TorresB()|TorresN()|DamasB()|DamasN(); }
inline Bitboard Tablero::AlfilesDamas(void) const { return AlfilesB()|AlfilesN()|DamasB()|DamasN(); }

// Versiones con referencias-privadas
inline Bitboard& Tablero::PeonesB(void) { return piezas[PeonB]; }
inline Bitboard& Tablero::PeonesN(void) { return piezas[PeonN]; }
inline Bitboard& Tablero::CaballosB(void) { return piezas[CaballoB]; }
inline Bitboard& Tablero::CaballosN(void) { return piezas[CaballoN]; }
inline Bitboard& Tablero::AlfilesB(void) { return piezas[AlfilB]; }
inline Bitboard& Tablero::AlfilesN(void) { return piezas[AlfilN]; }
inline Bitboard& Tablero::TorresB(void) { return piezas[TorreB]; }
inline Bitboard& Tablero::TorresN(void) { return piezas[TorreN]; }
inline Bitboard& Tablero::DamasB(void) { return piezas[DamaB]; }
inline Bitboard& Tablero::DamasN(void) { return piezas[DamaN]; }
inline Bitboard& Tablero::ReyesB(void) { return Bitboards::Mask[posReyB]; }
inline Bitboard& Tablero::ReyesN(void) { return Bitboards::Mask[posReyN]; }
